from PyQt6.QtWidgets import QWidget, QInputDialog
from .MicrosoftSecrets import MicrosoftSecrets
from PyQt6.QtCore import QCoreApplication
from .gui.LoginWindow import LoginWindow
from typing import cast, TYPE_CHECKING
import minecraft_launcher_lib
import traceback
import uuid
import json
import sys
import os


if TYPE_CHECKING:
    from .Environment import Environment


class AccountBase:
    def __init__(self) -> None:
        self.id = ""

    def getType(self) -> str:
        raise NotImplementedError()

    def getName(self) -> str:
        raise NotImplementedError()

    def getMinecraftUUID(self) -> str:
        raise NotImplementedError()

    def getAccessToken(self) -> str:
        raise NotImplementedError()

    def login(self, parent: QWidget | None) -> bool:
        raise NotImplementedError()

    def reload(self) -> bool:
        raise NotImplementedError()

    def getJsonData(self) -> dict:
        raise NotImplementedError()


class MicrosoftAccount(AccountBase):
    def __init__(self) -> None:
        super().__init__()

        self._name = ""
        self._uuid = ""
        self._accessToken = ""
        self._refreshToken = ""

    @classmethod
    def fromJsonData(cls, data: dict) -> "MicrosoftAccount":
        account = cls()

        account.id = data["id"]
        account._name = data["name"]
        account._uuid = data["uuid"]
        account._accessToken = data["accessToken"]
        account._refreshToken = data["refreshToken"]

        return account

    def getType(self) -> str:
        return "Microsoft"

    def getName(self) -> str:
        return self._name

    def getMinecraftUUID(self) -> str:
        return self._uuid

    def getAccessToken(self) -> str:
        return self._accessToken

    def login(self, parent: QWidget | None) -> bool:
        loginWindow = LoginWindow(parent)
        loginWindow.exec()
        accountData = loginWindow.getAccountData()

        if accountData is None:
            return False

        self._name = accountData["name"]
        self._uuid = accountData["uuid"]
        self._accessToken = accountData["accessToken"]
        self._refreshToken = accountData["refreshToken"]

        return True

    def reload(self) -> bool:
        secrets = MicrosoftSecrets.get_secrets()

        try:
            data = minecraft_launcher_lib.microsoft_account.complete_refresh(secrets.client_id, secrets.secret, secrets.redirect_url, self._refreshToken)

            self._name = data["name"]
            self._uuid = data["id"]
            self._accessToken = data["access_token"]
            self.refreshToken = data["refresh_token"]

            return True
        except minecraft_launcher_lib.exceptions.InvalidRefreshToken:
            return False
        except Exception:
            print(traceback.format_exc(), file=sys.stderr)
            return False

    def getJsonData(self) -> dict:
        return {
            "type": "microsoft",
            "id": self.id,
            "name": self._name,
            "uuid": self._uuid,
            "accessToken": self._accessToken,
            "refreshToken": self._refreshToken,
        }


class DummyAccount(AccountBase):
    def __init__(self) -> None:
        super().__init__()

        self._name = ""

    @classmethod
    def fromJsonData(cls, data: dict) -> "DummyAccount":
        account = cls()

        account.id = data["id"]
        account._name = data["name"]
        account._uuid = data["uuid"]

        return account

    def getType(self) -> str:
        return "Dummy"

    def getName(self) -> str:
        return self._name

    def getMinecraftUUID(self) -> str:
        return self._uuid

    def getAccessToken(self) -> str:
        return ""

    def login(self, parent: QWidget | None) -> bool:
        name, ok = QInputDialog.getText(parent, QCoreApplication.translate("AccountManager", "Enter name"), QCoreApplication.translate("AccountManager", "Please enter a name for the dummy account"))

        if not ok:
            return False

        self._name = name
        self._uuid = str(uuid.uuid4())

        return True

    def reload(self) -> bool:
        return True

    def getJsonData(self) -> dict:
        return {
            "type": "dummy",
            "id": self.id,
            "name": self._name,
            "uuid": self._uuid,
        }


class AccountManager:
    def __init__(self, env: "Environment") -> None:
        self._accountList: list[AccountBase] = []
        self._currentID = ""
        self._env = env

        self._loadData()

    def _loadData(self) -> None:
        if not os.path.isfile(os.path.join(self._env.dataDir, "accounts.json")):
            return

        try:
            with open(os.path.join(self._env.dataDir, "accounts.json"), "r", encoding="utf-8") as f:
                data = json.load(f)

            self._currentID = data["selectedAccount"]
            self._accountList.clear()

            for accountData in data["accounts"]:
                match accountData["type"]:
                    case "microsoft":
                        self._accountList.append(MicrosoftAccount.fromJsonData(accountData))
                    case "dummy":
                        self._accountList.append(DummyAccount.fromJsonData(accountData))

        except Exception:
            print(traceback.format_exc(), file=sys.stderr)

    def saveData(self) -> None:
        if self._env.args.dont_save_data:
            return

        try:
            data = {"selectedAccount": self._currentID, "accounts": []}

            for account in self._accountList:
                cast(list[dict], data["accounts"]).append(account.getJsonData())

            try:
                os.makedirs(self._env.dataDir)
            except FileExistsError:
                pass

            with open(os.path.join(self._env.dataDir, "accounts.json"), "w", encoding="utf-8") as f:
                json.dump(data, f, ensure_ascii=False, indent=4)
        except Exception:
            print(traceback.format_exc(), file=sys.stderr)

    def _generateID(self) -> str:
        while True:
            currentUUID = str(uuid.uuid4())
            if self.getAccountByID(currentUUID) is None:
                return currentUUID

    def getAccountByID(self, uuid: str) -> AccountBase | None:
        for currentAccount in self._accountList:
            if currentAccount.id == uuid:
                return currentAccount
        return None

    def getAccountByName(self, name: str) -> AccountBase | None:
        for currentAccount in self._accountList:
            if currentAccount.getName().lower() == name.lower():
                return currentAccount
        return None

    def getSelectedAccount(self) -> AccountBase | None:
        return self.getAccountByID(self._currentID)

    def getAllAccounts(self) -> list[AccountBase]:
        return self._accountList

    def setSelectedAccount(self, account: AccountBase) -> None:
        self._currentID = account.id
        self.saveData()

    def removeAccount(self, account: AccountBase) -> None:
        for count, currentAccount in enumerate(self._accountList):
            if account == currentAccount:
                del self._accountList[count]
                break

        if self._currentID == account.id:
            if len(self._accountList) == 0:
                self._currentID = ""
            else:
                self._currentID = self._accountList[0].id

        self.saveData()

    def _addAccount(self, account: AccountBase, parent: QWidget | None) -> AccountBase | None:
        if account.login(parent) is False:
            return None

        account.id = self._generateID()

        self._accountList.append(account)

        if self.getSelectedAccount() is None:
            self._currentID = account.id

        self.saveData()

        return account

    def addMicrosoftAccount(self, parent: QWidget | None) -> AccountBase | None:
        account = MicrosoftAccount()
        return self._addAccount(account, parent)

    def addDummyAccount(self, parent: QWidget | None) -> AccountBase | None:
        account = DummyAccount()
        return self._addAccount(account, parent)



from jdMinecraftLauncher.Functions import isFrozen
from PyQt6.QtCore import QCoreApplication, QTranslator
from PyQt6.QtWidgets import QWidget, QMessageBox
from typing import cast, TYPE_CHECKING
from PyQt6.QtGui import QIcon
from enum import Enum
import subprocess
import platform
import pathlib
import sys
import os


if TYPE_CHECKING:
    from jdMinecraftLauncher.Environment import Environment
    from jdMinecraftLauncher.Profile import Profile


class ShortcutLocation(Enum):
    DESKTOP = 0
    MENU = 1


def _createLinuxShortcut(env: "Environment", parent: QWidget | None, path: str, profile: "Profile") -> None:
    try:
        import desktop_entry_lib
    except ModuleNotFoundError:
        QMessageBox.critical(parent, QCoreApplication.translate("Shortcut", "desktop-entry-lib not found"), QCoreApplication.translate("Shortcut", "You need the desktop-entry-lib Python package to create a shortcut"))
        return

    entry = desktop_entry_lib.DesktopEntry()
    entry.Name.default_text = profile.name
    entry.Icon = "page.codeberg.JakobDev.jdMinecraftLauncher"
    entry.Exec = subprocess.list2cmdline(["xdg-open", "jdMinecraftLauncher:LaunchProfileByID/" + profile.id])
    entry.Categories.append("Game")
    entry.Comment.default_text = f"Start the {profile.name} profile in jdMinecraftLauncher"

    for langFile in os.listdir(os.path.join(env.currentDir, "translations")):
        if not langFile.endswith(".qm"):
            continue

        lang = langFile.removeprefix("jdMinecraftLauncher_").removesuffix(".qm")
        fullPath = os.path.join(env.currentDir, "translations", langFile)
        translator = QTranslator()
        translator.load(fullPath)
        comment = translator.translate("Shortcut", "Start the {{name}} profile in jdMinecraftLauncher").replace("{{name}}", profile.name)

        if comment != "":
            entry.Comment.translations[lang] = comment

    entry.write_file(os.path.join(path, f"page.codeberg.JakobDev.Profile.{profile.name}.desktop"))

    subprocess.run(["chmod", "+x", os.path.join(path, f"page.codeberg.JakobDev.Profile.{profile.name}.desktop")])


def _ensureWindowsUrlSchema(env: "Environment", parent: QWidget | None) -> None:
    if sys.platform != "win32":
        return
    import winreg

    # Check if the Schema already exists
    try:
        winreg.OpenKey(winreg.HKEY_CLASSES_ROOT, "jdMinecraftLauncher").Close()
        return
    except FileNotFoundError:
        pass

    if QMessageBox.question(parent, QCoreApplication.translate("Shortcut", "Add URL Schema"), QCoreApplication.translate("Shortcut", "To make Shortcuts work, you need to add the jdMinecraftLauncher URL Schema to Windows. Should it be added?"), QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No) != QMessageBox.StandardButton.Yes:
        return

    with winreg.CreateKeyEx(winreg.HKEY_CURRENT_USER, r"SOFTWARE\Classes\jdMinecraftLauncher", 0, winreg.KEY_WRITE) as protocolKey:
        winreg.SetValueEx(protocolKey, "URL Protocol", 0, winreg.REG_SZ, "")

    with winreg.CreateKeyEx(winreg.HKEY_CURRENT_USER, r"SOFTWARE\Classes\jdMinecraftLauncher\shell\open\command", 0, winreg.KEY_WRITE) as commandKey:
        if isFrozen():
            winreg.SetValueEx(commandKey, None, 0, winreg.REG_SZ, subprocess.list2cmdline([os.path.abspath(sys.argv[0]), "%1"]))
        else:
            winreg.SetValueEx(commandKey, None, 0, winreg.REG_SZ, subprocess.list2cmdline([sys.executable, os.path.abspath(sys.argv[0]), "%1"]))


def _createWindowsShortcut(path: str, profile: "Profile") -> None:
    try:
        os.makedirs(path)
    except Exception:
        pass

    with open(os.path.join(path, f"{profile.name}.url"), "w", encoding="utf-8", newline="\r\n") as f:
        f.write("[InternetShortcut]\n")
        f.write(f"URL=jdMinecraftLauncher:LaunchProfileByID/{profile.id}\n")


def canCreateShortcuts() -> bool:
    return platform.system() in ["Linux", "Windows"]


def createShortcut(env: "Environment", parent: QWidget | None, profile: "Profile", location: ShortcutLocation) -> None:
    if platform.system() == "Linux":
        if location == ShortcutLocation.DESKTOP:
            _createLinuxShortcut(env, parent, subprocess.check_output(["xdg-user-dir", "DESKTOP"]).decode("utf-8").strip(), profile)
        elif location == ShortcutLocation.MENU:
            _createLinuxShortcut(env, parent, os.path.expanduser("~/.local/share/applications"), profile)
    elif platform.system() == "Windows":
        _ensureWindowsUrlSchema(env, parent)
        if location == ShortcutLocation.DESKTOP:
            _createWindowsShortcut(str(pathlib.Path.home() / "Desktop"), profile)
        elif location == ShortcutLocation.MENU:
            _createWindowsShortcut(os.path.join(cast(str, os.getenv("APPDATA")), "Microsoft", "Windows", "Start Menu", "Programs"), profile)


def askCreateShortcut(env: "Environment", parent: QWidget | None, profile: "Profile") -> None:
    box = QMessageBox(parent)
    box.setText(QCoreApplication.translate("Shortcut", "Select where you want to create the Shortcut"))
    box.setWindowTitle(QCoreApplication.translate("Shortcut", "Create Shortcut"))
    box.setStandardButtons(QMessageBox.StandardButton.Save | QMessageBox.StandardButton.Discard | QMessageBox.StandardButton.Cancel)

    desktopButton = box.button(QMessageBox.StandardButton.Save)
    desktopButton.setText(QCoreApplication.translate("Shortcut", "Desktop"))
    desktopButton.setIcon(QIcon())

    menuButton = box.button(QMessageBox.StandardButton.Discard)
    menuButton.setText(QCoreApplication.translate("Shortcut", "Menu"))
    menuButton.setIcon(QIcon())

    bothButton = box.button(QMessageBox.StandardButton.Cancel)
    bothButton.setText(QCoreApplication.translate("Shortcut", "Both"))
    bothButton.setIcon(QIcon())

    box.exec()

    if box.clickedButton() == desktopButton:
        createShortcut(env, parent, profile, ShortcutLocation.DESKTOP)
    elif box.clickedButton() == menuButton:
        createShortcut(env, parent, profile, ShortcutLocation.MENU)
    elif box.clickedButton() == bothButton:
        createShortcut(env, parent, profile, ShortcutLocation.DESKTOP)
        createShortcut(env, parent, profile, ShortcutLocation.MENU)



# This file is used durring the build process
from setuptools import build_meta as origin
from typing import Optional, Any
import subprocess
import tempfile
import pathlib
import shutil
import sys
import os


TRANSLATION_DIRS = [
    "jdMinecraftLauncher/translations"
]


def get_lrelease_command() -> Optional[str]:
    for i in ("lrelease", "lrelease6", "pyside6-lrelease", "pyside5-lrelease"):
        if shutil.which(i) is not None:
            return i
    return None


def compile_ui(path: str) -> None:
    if shutil.which("pyuic6") is None:
        print("pyuic6 was not found", file=sys.stderr)
        sys.exit(1)

    program_dir = pathlib.Path(path) / "jdMinecraftLauncher"
    compiled_dir = program_dir / "ui_compiled"
    source_dir = program_dir / "ui"

    try:
        shutil.rmtree(compiled_dir)
    except FileNotFoundError:
        pass

    compiled_dir.mkdir()
    (compiled_dir / "__init__.py").touch()

    for i in source_dir.iterdir():
        if i.suffix != ".ui":
            continue

        subprocess.run(["pyuic6", str(i), "-o", str(compiled_dir / f"{i.stem}.py")], check=True)


def build_translations(path: str) -> None:
    command = get_lrelease_command()
    for i in TRANSLATION_DIRS:
        for ts in pathlib.Path(os.path.join(path, i)).iterdir():
            subprocess.run([command, str(ts), "-qm", os.path.join(path, i, ts.stem + ".qm")], check=True)
            os.remove(ts)


prepare_metadata_for_build_wheel = origin.prepare_metadata_for_build_wheel
get_requires_for_build_sdist = origin.get_requires_for_build_sdist
build_sdist = origin.build_sdist


def build_wheel(wheel_directory: str, config_settings: Any = None, metadata_directory: Optional[str] = None) -> str:
    if get_lrelease_command() is None:
        print("lrealease was not found", file=sys.stderr)
        sys.exit(1)

    wheel_name = origin.build_wheel(wheel_directory, config_settings=config_settings, metadata_directory=metadata_directory)
    wheel_path = os.path.join(wheel_directory, wheel_name)
    with tempfile.TemporaryDirectory() as tempdir:
        subprocess.run(["wheel", "unpack", "--dest", tempdir, wheel_path])
        current_dir = os.path.join(tempdir, os.listdir(tempdir)[0])

        compile_ui(current_dir)
        build_translations(current_dir)

        try:
            os.remove(os.path.join(current_dir, os.path.basename(__file__)))
        except FileNotFoundError:
            pass

        try:
            shutil.rmtree(os.path.join(current_dir, "jdMinecraftLauncher", "ui"))
        except FileNotFoundError:
            pass

        subprocess.run(["wheel", "pack", "--dest-dir", wheel_directory, current_dir])

    return wheel_name


def get_requires_for_build_wheel(self: Any, config_settings: Any = None) -> list[str]:
    if get_lrelease_command() is None:
        return origin.get_requires_for_build_wheel() + ["PySide6"]
    else:
        return origin.get_requires_for_build_wheel()



from PyQt6.QtWidgets import QTableWidget, QAbstractItemView, QHeaderView, QTableWidgetItem, QMenu, QMessageBox
from ...Shortcut import canCreateShortcuts, askCreateShortcut
from PyQt6.QtGui import QContextMenuEvent, QAction
from PyQt6.QtCore import Qt, QCoreApplication
from ...Functions import openFile
from typing import TYPE_CHECKING
from ...Profile import Profile


if TYPE_CHECKING:
    from ...Environment import Environment
    from .MainWindow import MainWindow


class ProfileEditorTab(QTableWidget):
    def __init__(self, env: "Environment", mainWindow: "MainWindow") -> None:
        super().__init__(0, 2)

        self._env = env
        self._mainWindow = mainWindow
        self._profileList: list["Profile"] = []

        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setHorizontalHeaderLabels((QCoreApplication.translate("ProfileEditorTab", "Profile Name"), QCoreApplication.translate("ProfileEditorTab", "Minecraft Version")))
        self.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
        self.horizontalHeader().setSectionResizeMode(1, QHeaderView.ResizeMode.Stretch)
        self.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        self.verticalHeader().hide()
        self.setShowGrid(False)

        self.updateProfiles()

    def updateProfiles(self) -> None:
        self._profileList.clear()
        while self.rowCount() > 0:
            self.removeRow(0)

        count = 0
        for i in self._env.profileCollection.profileList:
            nameItem = QTableWidgetItem(i.name)
            nameItem.setFlags(nameItem.flags() ^ Qt.ItemFlag.ItemIsEditable)
            if i.useLatestVersion:
                versionItem = QTableWidgetItem(QCoreApplication.translate("ProfileEditorTab", "(Latest version)"))
            elif i.useLatestSnapshot:
                versionItem = QTableWidgetItem(QCoreApplication.translate("ProfileEditorTab", "(Latest snapshot)"))
            else:
                versionItem = QTableWidgetItem(i.version)
            versionItem.setFlags(versionItem.flags() ^ Qt.ItemFlag.ItemIsEditable)
            self.insertRow(count)
            self.setItem(count, 0, nameItem)
            self.setItem(count, 1, versionItem)
            self._profileList.append(i)
            count += 1

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:  # type: ignore
        self.menu = QMenu(self)

        addProfile = QAction(QCoreApplication.translate("ProfileEditorTab", "Add Profile"), self)
        addProfile.triggered.connect(self._addProfile)
        self.menu.addAction(addProfile)

        editProfile = QAction(QCoreApplication.translate("ProfileEditorTab", "Edit Profile"), self)
        editProfile.triggered.connect(self._editProfile)
        self.menu.addAction(editProfile)

        copyProfile = QAction(QCoreApplication.translate("ProfileEditorTab", "Copy Profile"), self)
        copyProfile.triggered.connect(self._copyProfile)
        self.menu.addAction(copyProfile)

        removeProfile = QAction(QCoreApplication.translate("ProfileEditorTab", "Remove Profile"), self)
        removeProfile.triggered.connect(self._removeProfile)
        self.menu.addAction(removeProfile)

        openGameFolder = QAction(QCoreApplication.translate("ProfileEditorTab", "Open Game Folder"), self)
        openGameFolder.triggered.connect(lambda: openFile(self._env.profileCollection.profileList[self.currentRow()].getGameDirectoryPath()))
        self.menu.addAction(openGameFolder)

        if canCreateShortcuts():
            createShortcut = QAction(QCoreApplication.translate("ProfileEditorTab", "Create Shortcut"), self)
            createShortcut.triggered.connect(lambda: askCreateShortcut(self._env, self._mainWindow, self._env.profileCollection.profileList[self.currentRow()]))
            self.menu.addAction(createShortcut)

        self.menu.popup(event.globalPos())

    def _addProfile(self) -> None:
        self._mainWindow.profileWindow.loadProfile(Profile(QCoreApplication.translate("ProfileEditorTab", "New Profile"), self._env), True)
        self._mainWindow.profileWindow.open()

    def _editProfile(self) -> None:
        self._mainWindow.profileWindow.loadProfile(self._env.profileCollection.profileList[self.currentRow()], False)
        self._mainWindow.profileWindow.open()

    def _copyProfile(self) -> None:
        self._mainWindow.profileWindow.loadProfile(self._env.profileCollection.profileList[self.currentRow()], True, True)
        self._mainWindow.profileWindow.open()

    def _removeProfile(self) -> None:
        if len(self._env.profileCollection.profileList) == 1:
            QMessageBox.critical(self._mainWindow, QCoreApplication.translate("ProfileEditorTab", "Can't delete Profile"), QCoreApplication.translate("ProfileEditorTab", "You can't delete all Profiles. At least one Profile must stay."))
        else:
            self._env.profileCollection.removeProfileById(self._profileList[self.currentRow()].id)
            self._mainWindow.updateProfileList()



from .Constants import NewsTypeSetting, DisplayServerSetting
from typing import Optional, Any
import traceback
import json
import sys
import os


class Settings():
    def __init__(self) -> None:
        self._default_settings: dict[str, Any] = {
            "language": "default",
            "newsType": NewsTypeSetting.MINECRAFT,
            "newsFeedURL": "https://jakobdev.codeberg.page/blog/rss.xml",
            "newsFeedDefaultBrowser": True,
            "newsURL": "https://www.minecraft.net",
            "enableMultiLaunch": False,
            "extractNatives": False,
            "windowIconProgress": True,
            "customMinecraftDir": None,
            "useFlatpakSubsandbox": True,
            "checkUpdatesStartup": True,
            "displayServer": DisplayServerSetting.AUTO,
        }

        self._user_settings: dict[str, Any] = {}
        self._loadError: str | None = None

    def get(self, key: str) -> Any:
        """Returns the given setting"""
        if key in self._user_settings:
            return self._user_settings[key]
        elif key in self._default_settings:
            return self._default_settings[key]
        else:
            return None

    def set(self, key: str, value: Any) -> None:
        """Set the value of a setting"""
        self._user_settings[key] = value

    def save(self, path: str) -> None:
        """Save settings into file"""
        if len(self._user_settings) == 0 and not os.path.isfile(path):
            return

        try:
            with open(path, "w", encoding="utf-8") as f:
                json.dump(self._user_settings, f, ensure_ascii=False, indent=4)
        except Exception:
            print(traceback.format_exc(), file=sys.stderr)

    def load(self, path: str) -> None:
        """Load settings from file"""
        if not os.path.isfile(path):
            return

        try:
            with open(path, "r", encoding="utf-8") as f:
                self._user_settings = json.load(f)
        except Exception:
            self._loadError = traceback.format_exc()
            print(self._loadError, file=sys.stderr)

    def reset(self) -> None:
        """Resets all settings to the default values"""
        self._user_settings.clear()

    def getLoadError(self) -> Optional[str]:
        return self._loadError



from typing import TypedDict, TYPE_CHECKING
from ...Constants import NewsTypeSetting
import minecraft_launcher_lib
import requests
import base64
import html
import os


if TYPE_CHECKING:
    from ...Environment import Environment


class _NewsEntry(TypedDict):
    title: str
    summary: str
    link: str


def _newsListToHtml(newsList: list[_NewsEntry]) -> str:
    if len(newsList) == 0:
        return "<div>No News found</div>"

    html_code = ""

    for currentNews in newsList:
        html_code += '<div style="margin-top:16px">\n'
        html_code += f'<h3><a href="{currentNews["link"]}"">{currentNews["title"]}</a></h3>\n'
        html_code += f'<p class="content">{currentNews["summary"]}</p>\n'
        html_code += "</p></div>"

    return html_code


def _getMinecraftNews() -> list[_NewsEntry]:
    minecraftNews = minecraft_launcher_lib.news.get_minecraft_news()
    newsList = []

    for currentNews in minecraftNews["entries"]:
        newsList.append(_NewsEntry({
            "title": currentNews["title"],
            "summary": currentNews["text"],
            "link": currentNews["readMoreLink"],
        }))

    return newsList


def _parseRssFeed(env: "Environment") -> list[_NewsEntry]:
    import feedparser  # type:ignore[import-untyped]

    header = {"user-agent": f"jdMinecraftLauncher/{env.launcherVersion}"}
    r = requests.get(env.settings.get("newsFeedURL"), headers=header)
    feed = feedparser.parse(r.text)

    newsList = []

    for entry in feed["entries"]:
        if entry["title_detail"]["type"] == "text/html":
            title = entry["title"]
        else:
            title = html.escape(entry["title"])

        if entry["summary_detail"]["type"] == "text/html":
            summary = entry["summary"]
        else:
            summary = html.escape(entry["summary"])

        newsList.append(_NewsEntry({
            "title": title,
            "summary": summary,
            "link": entry["link"],
        }))

    return newsList


def GenerateNewsPage(env: "Environment") -> str:
    with open(os.path.join(env.currentDir, "utils", "NewsGenerator", "files", "News.html"), "r", encoding="utf-8") as f:
        template = f.read()

    with open(os.path.join(env.currentDir, "utils", "NewsGenerator", "files", "Background.png"), "rb") as f:
        template = template.replace("{{BackgroundImage}}", base64.b64encode(f.read()).decode("utf-8"))

    with open(os.path.join(env.currentDir, "utils", "NewsGenerator", "files", "Style.css"), "r", encoding="utf-8") as f:
        template = template.replace("{{Style}}", f.read())

    match env.settings.get("newsType"):
        case NewsTypeSetting.MINECRAFT:
            template = template.replace("{{ItemList}}", _newsListToHtml(_getMinecraftNews()))
        case NewsTypeSetting.RSS:
            template = template.replace("{{ItemList}}", _newsListToHtml(_parseRssFeed(env)))

    return template



from PyQt6.QtWidgets import QDialog, QWidget, QVBoxLayout, QMessageBox
from PyQt6.QtWebEngineWidgets import QWebEngineView
from PyQt6.QtCore import QUrl, QCoreApplication
from ..MicrosoftSecrets import MicrosoftSecrets
import minecraft_launcher_lib


class LoginWindow(QDialog):
    def __init__(self, parent: QWidget | None) -> None:
        super().__init__(parent)

        self._secrets = MicrosoftSecrets.get_secrets()

        self.setWindowTitle(QCoreApplication.translate("LoginWindow", "Login"))

        self._webView = QWebEngineView()

        loginUrl, self._state, self._codeVerifier = minecraft_launcher_lib.microsoft_account.get_secure_login_data(self._secrets.client_id, self._secrets.redirect_url)

        # Open the login url
        self._webView.load(QUrl(loginUrl))

        # Connects a function that is called when the url changed
        self._webView.urlChanged.connect(self.newUrl)

        self._accountData: dict[str, str] | None = None

        mainLayout = QVBoxLayout()
        mainLayout.addWidget(self._webView)

        mainLayout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(mainLayout)

    def newUrl(self, url: QUrl) -> None:
        # Check if the url contains the code
        if not minecraft_launcher_lib.microsoft_account.url_contains_auth_code(url.toString()):
            return

        # Get the code from the url
        authCode = minecraft_launcher_lib.microsoft_account.parse_auth_code_url(url.toString(), self._state)
        # Do the login
        try:
            accountInformation = minecraft_launcher_lib.microsoft_account.complete_login(self._secrets.client_id, self._secrets.secret, self._secrets.redirect_url, authCode, self._codeVerifier)
        except minecraft_launcher_lib.exceptions.AccountNotOwnMinecraft:
            self.hide()

            text = QCoreApplication.translate("LoginWindow", "Your account appears to not own Minecraft.")
            text += " " + QCoreApplication.translate("LoginWindow", "You need an account that owns Minecraft to use jdMinecraftLauncher.")
            text += " " + QCoreApplication.translate("LoginWindow", "If you've purchased Minecraft and still encounter this error, try logging in with the official launcher first.")
            text += " " + QCoreApplication.translate("LoginWindow", "If the error still persists, please write a bug report.")
            QMessageBox.critical(self, QCoreApplication.translate("LoginWindow", "Account does not own Minecraft"), text)

            loginUrl, self._state, self._codeVerifier = minecraft_launcher_lib.microsoft_account.get_secure_login_data(self._secrets.client_id, self._secrets.redirect_url)

            self._webView.load(QUrl(loginUrl))

            self.show()

            return

        self._accountData = {
            "name": accountInformation["name"],
            "accessToken": accountInformation["access_token"],
            "refreshToken": accountInformation["refresh_token"],
            "uuid": accountInformation["id"]
        }

        self.close()

    def getAccountData(self) -> dict[str, str] | None:
        return self._accountData



from PyQt6.QtWidgets import QWidget, QPushButton, QMenu, QTableWidgetItem, QMessageBox
from ...Functions import clearTableWidget, stretchTableWidgetColumnsSize
from PyQt6.QtCore import Qt, QCoreApplication, QPoint
from ...ui_compiled.AccountTab import Ui_AccountTab
from typing import TYPE_CHECKING
from PyQt6.QtGui import QAction


if TYPE_CHECKING:
    from ...AccountManager import AccountBase
    from ...Environment import Environment
    from .MainWindow import MainWindow


class _TableColumns:
    NAME = 0
    TYPE = 1
    SWITCH = 2
    REMOVE = 3


class _SwitchAccountButton(QPushButton):
    def __init__(self, env: "Environment", account: "AccountBase", mainWindow: "MainWindow") -> None:
        super().__init__(QCoreApplication.translate("AccountTab", "Switch"))

        self._env = env
        self._account = account
        self._mainWindow = mainWindow

        self.clicked.connect(self._buttonClicked)

    def _buttonClicked(self) -> None:
        if not self._account.reload():
            if not self._account.login(self._mainWindow):
                return

        self._env.accountManager.setSelectedAccount(self._account)
        self._mainWindow.updateAccountInformation()


class _RemoveAccountButton(QPushButton):
    def __init__(self, env: "Environment", account: "AccountBase", mainWindow: "MainWindow") -> None:
        super().__init__(QCoreApplication.translate("AccountTab", "Remove"))

        self._env = env
        self._account = account
        self._mainWindow = mainWindow

        self.clicked.connect(self._buttonClicked)

    def _buttonClicked(self) -> None:
        if self._env.offlineMode:
            QMessageBox.critical(self, QCoreApplication.translate("AccountTab", "No Internet Connection"), QCoreApplication.translate("AccountTab", "This Feature needs a internet connection"))
            return

        self._env.accountManager.removeAccount(self._account)

        if self._env.accountManager.getSelectedAccount() is None:
            self._mainWindow.hide()
            if self._env.accountManager.addMicrosoftAccount(self) is None:
                self._mainWindow.close()
                return
            self._mainWindow.show()

        self._mainWindow.updateAccountInformation()


class AccountTab(QWidget, Ui_AccountTab):
    def __init__(self, env: "Environment", mainWindow: "MainWindow") -> None:
        super().__init__()

        self.setupUi(self)

        self._env = env
        self._mainWindow = mainWindow

        stretchTableWidgetColumnsSize(self.accountTable)
        self.accountTable.setColumnHidden(_TableColumns.TYPE, not env.debugMode)
        self.accountTable.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)

        self.accountTable.customContextMenuRequested.connect(self._openContextMenu)

        self.addButton.clicked.connect(self._addMicrosoftAccount)

    def _openContextMenu(self, pos: QPoint) -> None:
        menu = QMenu(self)

        addMicrosoftAccount = QAction(QCoreApplication.translate("AccountTab", "Add Account"), self)
        addMicrosoftAccount.triggered.connect(self._addMicrosoftAccount)
        menu.addAction(addMicrosoftAccount)

        if self._env.debugMode:
            addDummyAccount = QAction(QCoreApplication.translate("AccountTab", "Add Dummy Account"), self)
            addDummyAccount.triggered.connect(self._addDummyAccount)
            menu.addAction(addDummyAccount)

        menu.popup(self.accountTable.mapToGlobal(pos))

    def _addMicrosoftAccount(self) -> None:
        account = self._env.accountManager.addMicrosoftAccount(self)

        if account is None:
            return

        self._env.accountManager.setSelectedAccount(account)
        self._mainWindow.updateAccountInformation()

    def _addDummyAccount(self) -> None:
        account = self._env.accountManager.addDummyAccount(self)

        if account is None:
            return

        self._env.accountManager.setSelectedAccount(account)
        self._mainWindow.updateAccountInformation()

    def updateAccountTable(self) -> None:
        clearTableWidget(self.accountTable)

        for account in self._env.accountManager.getAllAccounts():
            row = self.accountTable.rowCount()
            self.accountTable.insertRow(row)

            nameItem = QTableWidgetItem(account.getName())
            nameItem.setFlags(nameItem.flags() ^ Qt.ItemFlag.ItemIsEditable)
            self.accountTable.setItem(row, _TableColumns.NAME, nameItem)

            typeItem = QTableWidgetItem(account.getType())
            typeItem.setFlags(typeItem.flags() ^ Qt.ItemFlag.ItemIsEditable)
            self.accountTable.setItem(row, _TableColumns.TYPE, typeItem)

            self.accountTable.setCellWidget(row, _TableColumns.SWITCH, _SwitchAccountButton(self._env, account, self._mainWindow))

            self.accountTable.setCellWidget(row, _TableColumns.REMOVE, _RemoveAccountButton(self._env, account, self._mainWindow))



from jdMinecraftLauncher.utils.WindowIconProgress import createWindowIconProgress
from ...ui_compiled.MainWindow import Ui_MainWindow
from PyQt6.QtWebEngineCore import QWebEngineProfile
from PyQt6.QtCore import QCoreApplication, QEvent
from PyQt6.QtWidgets import QWidget, QMessageBox
from ...RunMinecraft import getMinecraftCommand
from .ProfileEditorTab import ProfileEditorTab
from .VersionEditorTab import VersionEditorTab
from ...InstallThread import InstallThread
from ...AccountManager import AccountBase
from ..ProfileWindow import ProfileWindow
from .GameOutputTab import GameOutputTab
from typing import cast, TYPE_CHECKING
from PyQt6.QtGui import QCloseEvent
from .OptionsTab import OptionsTab
from .AccountTab import AccountTab
from .FabricTab import FabricTab
from .ForgeTab import ForgeTab
from .AboutTab import AboutTab
from .NewsTab import NewsTab
import urllib.parse
import tempfile
import platform
import sys
import os


if TYPE_CHECKING:
    from ...Environment import Environment
    from ...Profile import Profile


class MainWindow(QWidget, Ui_MainWindow):
    def __init__(self, env: "Environment") -> None:
        super().__init__()

        self.setupUi(self)

        self._env = env
        self.profileListRebuild = False
        self.profileWindow = ProfileWindow(self._env, self)
        self.windowIconProgress = createWindowIconProgress(self)

        self.windowIconProgress.hide()

        self.tabWidget.clear()

        QWebEngineProfile.defaultProfile().setHttpAcceptLanguage(env.locale.name())
        QWebEngineProfile.defaultProfile().setHttpUserAgent("jdMinecraftLauncher/" + env.launcherVersion)

        self._newsTab = NewsTab(env)
        self._newsTab.updateNews()

        self._profileEditorTab = ProfileEditorTab(env, self)
        self._versionEditorTab = VersionEditorTab(env)
        self._optionsTab = OptionsTab(env, self)
        self._forgeTab = ForgeTab(env, self)
        self._fabricTab = FabricTab(env, self)
        self._accountTab = AccountTab(env, self)
        self._aboutTab = AboutTab(env)

        self.tabWidget.addTab(self._newsTab, QCoreApplication.translate("MainWindow", "News"))
        self.tabWidget.addTab(self._profileEditorTab, QCoreApplication.translate("MainWindow", "Profile Editor"))
        self.tabWidget.addTab(self._versionEditorTab, QCoreApplication.translate("MainWindow", "Version Editor"))
        self.tabWidget.addTab(self._optionsTab, QCoreApplication.translate("MainWindow", "Options"))
        self.tabWidget.addTab(self._forgeTab, QCoreApplication.translate("MainWindow", "Forge"))
        self.tabWidget.addTab(self._fabricTab, QCoreApplication.translate("MainWindow", "Fabric"))
        self.tabWidget.addTab(self._accountTab, QCoreApplication.translate("MainWindow", "Accounts"))
        self.tabWidget.addTab(self._aboutTab, QCoreApplication.translate("MainWindow", "About"))

        self.newProfileButton.clicked.connect(self.newProfileButtonClicked)
        self.editProfileButton.clicked.connect(self.editProfileButtonClicked)
        self.playButton.clicked.connect(self.playButtonClicked)
        self.accountButton.clicked.connect(self.logoutButtonClicked)
        self.profileComboBox.currentIndexChanged.connect(self.profileComboBoxIndexChanged)

        self.updateProfileList()

        self.installThread = InstallThread(env)
        self.installThread.text.connect(lambda text: self.progressBar.setFormat(text))
        self.installThread.progress.connect(self.updateProgress)
        self.installThread.progress_max.connect(lambda progress_max: self.progressBar.setMaximum(progress_max))
        self.installThread.finished.connect(self.installFinish)

        self._is_first_open = False

    def openMainWindow(self) -> None:
        if self._is_first_open:
            self.show()
            return

        if platform.system() == "Linux":
            from jdMinecraftLauncher.DBusService import DBusService
            DBusService(self._env, self._env.app, self)

        if self._env.args.launch_profile:
            profile = self._env.profileCollection.getProfileByName(self._env.args.launch_profile)
            if profile:
                self.launchProfile(profile)
            else:
                QMessageBox.critical(self, QCoreApplication.translate("MainWindow", "Profile not found"), QCoreApplication.translate("MainWindow", "The given Profile was not found"))
        elif self._env.args.url:
            parse_results = urllib.parse.urlparse(self._env.args.url)
            if parse_results.scheme == "jdminecraftlauncher":
                self._handleCustomURL(parse_results.path)

        self._is_first_open = True
        self.show()

        if (loadError := self._env.profileCollection.getLoadError()) is not None:
            msgBox = QMessageBox()
            msgBox.setWindowTitle(QCoreApplication.translate("MainWindow", "Unable to load Profiles"))
            msgBox.setText(QCoreApplication.translate("MainWindow", "jdMinecraft was unable to load your profiles due to an error. Apologies for any inconvenience. Please report this bug."))
            msgBox.setDetailedText(loadError)
            msgBox.exec()

        if (loadError := self._env.settings.getLoadError()) is not None:
            msgBox = QMessageBox()
            msgBox.setWindowTitle(QCoreApplication.translate("MainWindow", "Unable to load Settings"))
            msgBox.setText(QCoreApplication.translate("MainWindow", "jdMinecraft was unable to load your settings due to an error. Apologies for any inconvenience. Please report this bug."))
            msgBox.setDetailedText(loadError)
            msgBox.exec()

    def updateNewsTab(self) -> None:
        self._newsTab.updateNews()

    def _handleCustomURL(self, args: str) -> None:
        try:
            method, param = args.split("/", 1)
        except ValueError:
            return

        if method == "LaunchProfileByID":
            profile = self._env.profileCollection.getProfileByID(param)
            if profile:
                self.launchProfile(profile)
            else:
                QMessageBox.critical(self, QCoreApplication.translate("MainWindow", "Profile not found"), QCoreApplication.translate("MainWindow", "The given Profile was not found"))
        elif method == "LaunchProfileByName":
            profile = self._env.profileCollection.getProfileByName(param)
            if profile:
                self.launchProfile(profile)
            else:
                QMessageBox.critical(self, QCoreApplication.translate("MainWindow", "Profile not found"), QCoreApplication.translate("MainWindow", "The given Profile was not found"))

    def updateProfileList(self) -> None:
        currentIndex = 0
        self.profileListRebuild = True
        self.profileComboBox.clear()

        for count, i in enumerate(self._env.profileCollection.profileList):
            self.profileComboBox.addItem(i.name)
            if i.id == self._env.profileCollection.selectedProfile:
                currentIndex = count

        self._profileEditorTab.updateProfiles()
        self.profileComboBox.setCurrentIndex(currentIndex)
        self.profileListRebuild = False

    def profileComboBoxIndexChanged(self, index: int) -> None:
        if not self.profileListRebuild:
            self._env.profileCollection.selectedProfile = self._env.profileCollection.profileList[index].id

    def newProfileButtonClicked(self) -> None:
        self.profileWindow.loadProfile(self._env.profileCollection.getSelectedProfile(), True, True)
        self.profileWindow.open()

    def editProfileButtonClicked(self) -> None:
        self.profileWindow.loadProfile(self._env.profileCollection.getSelectedProfile(), False)
        self.profileWindow.open()

    def launchProfile(self, profile: "Profile") -> None:
        if self._env.offlineMode:
            if os.path.isdir(os.path.join(self._env.minecraftDir, "versions", profile.getVersionID())):
                self.startMinecraft(profile)
            else:
                QMessageBox.critical(self, QCoreApplication.translate("MainWindow", "No Internet Connection"), QCoreApplication.translate("MainWindow", "You need a internet connection to install a new version, but you are still able to play already installed versions."))
        else:
            self.installVersion(profile)

    def playButtonClicked(self) -> None:
        self.launchProfile(self._env.profileCollection.getSelectedProfile())

    def logoutButtonClicked(self) -> None:
        if self._env.offlineMode:
            QMessageBox.critical(self, QCoreApplication.translate("MainWindow", "No Internet Connection"), QCoreApplication.translate("MainWindow", "This Feature needs a internet connection"))
            return

        self._env.accountManager.removeAccount(cast(AccountBase, self._env.accountManager.getSelectedAccount()))

        if self._env.accountManager.getSelectedAccount() is None:
            self.hide()
            if self._env.accountManager.addMicrosoftAccount(self) is None:
                self.close()
                return
            self.show()

        self.updateAccountInformation()

    def startMinecraft(self, profile: "Profile") -> None:
        if self._env.settings.get("extractNatives"):
            natives_path = tempfile.mktemp(prefix="minecraft_natives_")
        else:
            natives_path = ""

        args = getMinecraftCommand(self._env.profileCollection.getSelectedProfile(), self._env, natives_path)
        outputTab = GameOutputTab(self._env, self)
        tabID = self.tabWidget.addTab(outputTab, QCoreApplication.translate("MainWindow", "Game Output"))
        self.tabWidget.setCurrentIndex(tabID)
        outputTab.executeCommand(profile, args, natives_path)

    def installFinish(self) -> None:
        self.windowIconProgress.hide()

        if self.installThread.isVersionNotFound():
            QMessageBox.critical(self, QCoreApplication.translate("MainWindow", "Version not found"), QCoreApplication.translate("MainWindow", "The version used by this profile was not found and could not be installed. Perhaps you have uninstalled it."))
            return

        if self.installThread.getError() is not None:
            text = QCoreApplication.translate("MainWindow", "Due to an error, the installation could not be completed") + "<br><br>"
            text += QCoreApplication.translate("MainWindow", "This may have been caused by a network error")

            msgBox = QMessageBox()
            msgBox.setWindowTitle(QCoreApplication.translate("MainWindow", "Installation failed"))
            msgBox.setText(text)
            msgBox.setDetailedText(self.installThread.getError())
            msgBox.exec()

            self.progressBar.setValue(0)
            self.progressBar.setFormat("")
            self.setInstallButtonsEnabled(True)

            return

        if self.installThread.shouldStartMinecraft() and self.installThread.getError() is None:
            self._env.updateInstalledVersions()
            self._versionEditorTab.updateVersions()
            self.startMinecraft(self._current_running_profile)
        else:
            self._env.loadVersions()
            self.profileWindow.updateVersionsList()
            self.setInstallButtonsEnabled(True)

    def installVersion(self, profile: "Profile") -> None:
        self._current_running_profile = profile
        self.playButton.setEnabled(False)
        self.installThread.setup(profile)
        self.installThread.start()

    def updateAccountInformation(self) -> None:
        self.accountLabel.setText(QCoreApplication.translate("MainWindow", "Welcome, {{name}}").replace("{{name}}", self._env.accountManager.getSelectedAccount().getName()))
        self._profileEditorTab.updateProfiles()

        if self._env.offlineMode:
            self.playButton.setText(QCoreApplication.translate("MainWindow", "Play Offline"))
        else:
            self.playButton.setText(QCoreApplication.translate("MainWindow", "Play"))

        self._accountTab.updateAccountTable()

    def setInstallButtonsEnabled(self, enabled: bool) -> None:
        self.playButton.setEnabled(enabled)
        self._forgeTab.setButtonsEnabled(enabled)
        self._fabricTab.setButtonsEnabled(enabled)

    def updateProgress(self, progress: int) -> None:
        self.progressBar.setValue(progress)

        if self._env.settings.get("windowIconProgress"):
            self.windowIconProgress.setProgress(progress / self.progressBar.maximum())

    def event(self, event: QEvent) -> bool:  # type: ignore
        if event.type() == QEvent.Type.WinIdChange:
            self.windowIconProgress = createWindowIconProgress(self)

        return super().event(event)

    def closeEvent(self, event: QCloseEvent) -> None:  # type: ignore
        if self._env.args.dont_save_data:
            event.accept()
            sys.exit(0)

        self._env.profileCollection.save()
        self._optionsTab.saveSettings()

        event.accept()
        sys.exit(0)



from PyQt6.QtWidgets import QTableWidget, QAbstractItemView, QMenu, QHeaderView, QTableWidgetItem
from PyQt6.QtGui import QAction, QContextMenuEvent
from PyQt6.QtCore import Qt, QCoreApplication
from typing import TYPE_CHECKING
import shutil
import os


if TYPE_CHECKING:
    from ...Environment import Environment


class VersionEditorTab(QTableWidget):
    def __init__(self, env: "Environment") -> None:
        super().__init__(0, 2)
        self._env = env

        self._uninstallVersion = QAction(QCoreApplication.translate("VersionEditorTab", "Uninstall Version"), self)
        self._uninstallVersion.triggered.connect(self._uninstallVersionClicked)

        self.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self.setHorizontalHeaderLabels((QCoreApplication.translate("VersionEditorTab", "Minecraft Version"), QCoreApplication.translate("VersionEditorTab", "Version Type")))
        self.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
        self.horizontalHeader().setSectionResizeMode(1, QHeaderView.ResizeMode.Stretch)
        self.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        self.verticalHeader().hide()
        self.setShowGrid(False)

        self.updateVersions()

    def updateVersions(self) -> None:
        if len(self._env.installedVersion) == 0:
            self._uninstallVersion.setEnabled(False)
        else:
            self._uninstallVersion.setEnabled(True)

        while self.rowCount() > 0:
            self.removeRow(0)

        count = 0
        for i in self._env.installedVersion:
            idItem = QTableWidgetItem(i["id"])
            idItem.setFlags(idItem.flags() ^ Qt.ItemFlag.ItemIsEditable)
            typeItem = QTableWidgetItem(i["type"])
            typeItem.setFlags(typeItem.flags() ^ Qt.ItemFlag.ItemIsEditable)
            self.insertRow(count)
            self.setItem(count, 0, idItem)
            self.setItem(count, 1, typeItem)
            count += 1

    def _uninstallVersionClicked(self) -> None:
        shutil.rmtree(os.path.join(self._env.minecraftDir, "versions", self._env.installedVersion[self.currentRow()]["id"]))
        del self._env.installedVersion[self.currentRow()]
        self._env.updateInstalledVersions()
        self.updateVersions()

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:  # type: ignore
        self.menu = QMenu(self)

        self.menu.addAction(self._uninstallVersion)

        self.menu.popup(event.globalPos())



from PyQt6.QtWidgets import QPlainTextEdit, QMessageBox
from PyQt6.QtCore import QProcess, QCoreApplication
from typing import TYPE_CHECKING
import shutil
import sys


if TYPE_CHECKING:
    from ...Environment import Environment
    from .MainWindow import MainWindow
    from ...Profile import Profile


class GameOutputTab(QPlainTextEdit):
    def __init__(self, env: "Environment", mainWindow: "MainWindow") -> None:
        super().__init__()

        self._env = env
        self._mainWindow = mainWindow

        self.setLineWrapMode(QPlainTextEdit.LineWrapMode.NoWrap)
        self.setReadOnly(True)

    def dataReady(self) -> None:
        cursor = self.textCursor()
        cursor.movePosition(cursor.MoveOperation.End)
        cursor.insertText(bytes(self.process.readAll()).decode(encoding=sys.stdout.encoding, errors="replace"))  # type: ignore
        self.moveCursor(cursor.MoveOperation.End)

    def procStarted(self) -> None:
        if self.profile.launcherVisibility != 2:
            self._mainWindow.hide()
        self._mainWindow.playButton.setEnabled(self._env.settings.get("enableMultiLaunch"))

    def procFinish(self) -> None:
        if self.profile.launcherVisibility == 0:
            self._mainWindow.show()
            self._mainWindow.setFocus()
        elif self.profile.launcherVisibility == 1:
            self._mainWindow.close()

        self._mainWindow.playButton.setEnabled(True)

        if self.natives_path != "":
            try:
                shutil.rmtree(self.natives_path)
            except Exception:
                pass

    def executeCommand(self, profile: "Profile", command: list[str], natives_path: str) -> None:
        self.profile = profile
        self.natives_path = natives_path
        self.process = QProcess(self)
        self.process.setProcessChannelMode(QProcess.ProcessChannelMode.MergedChannels)
        self.process.setWorkingDirectory(profile.getGameDirectoryPath())
        self.process.readyRead.connect(self.dataReady)
        self.process.started.connect(self.procStarted)
        self.process.finished.connect(self.procFinish)
        self.process.start(command[0], command[1:])

        if not self.process.waitForStarted():
            self.setPlainText(QCoreApplication.translate("GameOutputTab", "Failed to start Minecraft"))
            QMessageBox.critical(self, QCoreApplication.translate("GameOutputTab", "Failed to start"), QCoreApplication.translate("GameOutputTab", "Minecraft could not be started. Maybe you use a invalid Java executable."))
            self._mainWindow.playButton.setEnabled(True)
            return

        # There are some cases e.g. when failing to init a window where Minecraft uses tinyfiledialogs to output an error message
        # When DISPLAY is not set (e.g. on Wayland mode) it writes to the terminal and waits for the user to press enter
        # So we simulate a enter press here, so it does not hang, as this is not a TTY which supports a key press
        self.process.write(b"\n")
        self.process.closeWriteChannel()



from PyQt6.QtCore import QCoreApplication
from PyQt6.QtWidgets import QMessageBox
from typing import TYPE_CHECKING
import webbrowser
import traceback
import requests
import sys


if TYPE_CHECKING:
    from ..Environment import Environment


def _checkUpdatesInternal(env: "Environment") -> None:
    latestVersion = requests.get("https://codeberg.org/api/v1/repos/JakobDev/jdMinecraftLauncher/releases/latest").json()["name"]

    if env.launcherVersion == latestVersion:
        return

    sourceforgeURL = f"https://sourceforge.net/projects/jdminecraftlauncher/files/{latestVersion}"

    if requests.head(sourceforgeURL).status_code != 200:
        return

    text = QCoreApplication.translate("UpdateChecker", "Version {{version}} of jdMinecraftLauncher is now available.").replace("{{version}}", latestVersion) + " "
    text += QCoreApplication.translate("UpdateChecker", "You are currently using version {{version}}.").replace("{{version}}", latestVersion) + " "
    text += QCoreApplication.translate("UpdateChecker", "Do you want to download the latest version?")

    if QMessageBox.question(None, QCoreApplication.translate("UpdateChecker", "New version available"), text, QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No) != QMessageBox.StandardButton.Yes:
        return

    webbrowser.open(sourceforgeURL)
    sys.exit(0)


def checkUpdates(env: "Environment") -> None:
    try:
        _checkUpdatesInternal(env)
    except Exception:
        print(traceback.format_exc(), file=sys.stderr)



from typing import TYPE_CHECKING
import traceback
import platform
import sys


if TYPE_CHECKING:
    from jdMinecraftLauncher.gui.MainWindow.MainWindow import MainWindow


class WindowIconProgressBase:
    def setProgress(self, value: float) -> None:
        pass

    def hide(self) -> None:
        pass

    def isSupported(self) -> bool:
        return False


class WindowIconProgressWindows(WindowIconProgressBase):
    def __init__(self, windowID: int) -> None:
        import PyTaskbar  # type:ignore[import-not-found]

        self._progress = PyTaskbar.Progress(windowID)
        self._progress.init()
        self._progress.setState("normal")

    def setProgress(self, value: float) -> None:
        self._progress.setProgress(int(value * 100))

    def hide(self) -> None:
        self._progress.setProgress(0)

    def isSupported(self) -> bool:
        return True


class WindowIconProgressUnix(WindowIconProgressBase):
    def __init__(self) -> None:
        from PyQt6.QtDBus import QDBusConnection, QDBusMessage

        self._connection = QDBusConnection.sessionBus()
        self._message = QDBusMessage.createSignal("/", "com.canonical.Unity.LauncherEntry", "Update")

    def setProgress(self, value: float) -> None:
        self._message.setArguments(("application://page.codeberg.JakobDev.jdMinecraftLauncher.desktop", {"progress": value, "progress-visible": True}))
        self._connection.send(self._message)

    def hide(self) -> None:
        self._message.setArguments(("application://page.codeberg.JakobDev.jdMinecraftLauncher.desktop", {"progress-visible": False}))
        self._connection.send(self._message)

    def isSupported(self) -> bool:
        return True


def createWindowIconProgress(mainWindow: "MainWindow") -> WindowIconProgressBase:
    try:
        if platform.system() == "Windows":
            try:
                return WindowIconProgressWindows(int(mainWindow.winId()))
            except ModuleNotFoundError:
                print("Could not import PyTaskbar", file=sys.stderr)
                return WindowIconProgressBase()
        elif platform.system() == "Linux":
            return WindowIconProgressUnix()
        else:
            return WindowIconProgressBase()
    except Exception:
        print("Could not create WindowIconProgress", file=sys.stderr)
        print(traceback.format_exc(), end="", file=sys.stderr)
        return WindowIconProgressBase()



from PyQt6.QtCore import QThread, pyqtSignal
from typing import Optional, TYPE_CHECKING
import minecraft_launcher_lib
import traceback


if TYPE_CHECKING:
    from .Environment import Environment
    from .Profile import Profile


class InstallThread(QThread):
    progress_max = pyqtSignal("int")
    progress = pyqtSignal("int")
    text = pyqtSignal("QString")

    def __init__(self, env: "Environment") -> None:
        QThread.__init__(self)
        self.callback: minecraft_launcher_lib.types.CallbackDict = {
            "setStatus": lambda text: self.text.emit(text),
            "setMax": lambda max_progress: self.progress_max.emit(max_progress),
            "setProgress": lambda progress: self.progress.emit(progress),
        }
        self.forgeVersion: str | None = None
        self.fabricVersion: str | None = None
        self._currentError: str | None = None
        self._versionNotFound = False
        self.env = env

    def __del__(self) -> None:
        self.wait()

    def _setupAll(self) -> None:
        self.forgeVersion = None
        self.fabricVersion = None
        self._currentError = None
        self._versionNotFound = False

    def setup(self, profile: "Profile") -> None:
        self._setupAll()
        self.profile = profile
        self.startMinecraft = True

    def setupForgeInstallation(self, forgeVersion: str) -> None:
        self._setupAll()
        self.forgeVersion = forgeVersion
        self.startMinecraft = False

    def setupFabricInstallation(self, fabricVersion: str) -> None:
        self._setupAll()
        self.fabricVersion = fabricVersion
        self.startMinecraft = False

    def shouldStartMinecraft(self) -> bool:
        return self.startMinecraft

    def getError(self) -> Optional[str]:
        return self._currentError

    def isVersionNotFound(self) -> bool:
        return self._versionNotFound

    def run(self) -> None:
        try:
            if self.forgeVersion:
                minecraft_launcher_lib.forge.install_forge_version(self.forgeVersion, self.env.minecraftDir, callback=self.callback)
            elif self.fabricVersion:
                minecraft_launcher_lib.fabric.install_fabric(self.fabricVersion, self.env.minecraftDir, callback=self.callback)
            else:
                minecraft_launcher_lib.install.install_minecraft_version(self.profile.getVersionID(), self.env.minecraftDir, callback=self.callback)
        except minecraft_launcher_lib.exceptions.VersionNotFound:
            self._versionNotFound = True
        except Exception:
            self._currentError = traceback.format_exc()



from jdMinecraftLauncher.Profile import Profile
from PyQt6.QtCore import QCoreApplication
from PyQt6.QtWidgets import QMessageBox
from typing import TYPE_CHECKING
import minecraft_launcher_lib
import traceback


if TYPE_CHECKING:
    from jdMinecraftLauncher.gui.MainWindow.MainWindow import MainWindow
    from jdMinecraftLauncher.Environment import Environment


def _convertProfile(env: "Environment", vanilla_profile: minecraft_launcher_lib.types.VanillaLauncherProfile) -> Profile:
    profile = Profile(vanilla_profile["name"], env)

    profile.useLatestVersion = False
    if vanilla_profile["versionType"] == "latest-release":
        profile.useLatestVersion = True
    elif vanilla_profile["versionType"] == "latest-snapshot":
        profile.useLatestSnapshot = True
        profile.enableSnapshots = True
    else:
        profile.version = vanilla_profile["version"]  # type: ignore
        for i in env.versions["versions"]:
            if i["id"] == vanilla_profile["version"]:
                if i["type"] == "old_alpha":
                    profile.enableAlpha = True
                elif i["type"] == "old_beta":
                    profile.enableBeta = True
                elif i["type"] == "snapshot":
                    profile.enableSnapshots = True

    if vanilla_profile.get("gameDirectory") is not None:
        profile.gameDirectoryPath = vanilla_profile["gameDirectory"]
        profile.customGameDirectory = True

    if vanilla_profile.get("customResolution") is not None:
        profile.resolutionY = str(vanilla_profile["customResolution"]["height"])  # type: ignore
        profile.resolutionY = str(vanilla_profile["customResolution"]["width"])  # type: ignore
        profile.customResolution = True

    if vanilla_profile.get("javaExecutable") is not None:
        profile.executable = vanilla_profile["javaExecutable"]  # type: ignore
        profile.customExecutable = True

    if vanilla_profile.get("javaArguments") is not None:
        profile.arguments = " ".join(vanilla_profile["javaArguments"])  # type: ignore
        profile.customArguments = True

    return profile


def _importProfiles(env: "Environment") -> list[Profile]:
    profileList = []
    for i in minecraft_launcher_lib.vanilla_launcher.load_vanilla_launcher_profiles(env.minecraftDir):
        profileList.append(_convertProfile(env, i))

    return profileList


def askProfileImport(env: "Environment", mainWindow: "MainWindow") -> None:
    if not minecraft_launcher_lib.vanilla_launcher.do_vanilla_launcher_profiles_exists(env.minecraftDir):
        return

    if QMessageBox.question(mainWindow, QCoreApplication.translate("ProfileImporter", "Import Profiles"), QCoreApplication.translate("ProfileImporter", "jdMinecraftLauncher can import Profiles from the vanilla Launcher. Do you want to import your Profiles?"), QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No) != QMessageBox.StandardButton.Yes:
        return

    try:
        profiles = _importProfiles(env)
    except Exception:
        msgBox = QMessageBox()
        msgBox.setWindowTitle(QCoreApplication.translate("ProfileImporter", "Error"))
        msgBox.setText(QCoreApplication.translate("ProfileImporter", "Due to an error, the profiles could not be imported. Sorry for that."))
        msgBox.setDetailedText(traceback.format_exc())
        msgBox.exec()
        return

    for i in profiles:
        env.profileCollection.profileList.append(i)

    mainWindow.updateProfileList()



from PyQt6.QtWidgets import QTableWidget, QHeaderView, QComboBox, QApplication
from typing import Dict, List, Any, TYPE_CHECKING
import subprocess
import platform
import requests
import shutil
import socket
import json
import sys
import os


if TYPE_CHECKING:
    from jdMinecraftLauncher.Environment import Environment


def openFile(path: str) -> None:
    if platform.system() == "Windows":
        os.startfile(path)  # type: ignore
    elif platform.system() == "Darwin":
        subprocess.Popen(["open", path])
    else:
        subprocess.Popen(["xdg-open", path])


def saveProfiles(env: "Environment") -> None:
    if env.args.dont_save_data:
        return

    profileList = []
    for a in env.profiles:
        b = vars(a)
        c = {}
        for key, value in b.items():
            if value != env:
                c[key] = value
        profileList.append(c)
    with open(os.path.join(env.dataDir, "profiles.json"), 'w', encoding='utf-8') as f:
        json.dump({"selectedProfile": env.selectedProfile, "profileList": profileList}, f, ensure_ascii=False, indent=4)


def hasInternetConnection() -> bool:
    try:
        socket.create_connection(("api.mojang.com", 80))
        return True
    except OSError:
        return False


def downloadFile(url: str, path: str) -> None:
    if os.path.isfile(path):
        return
    try:
        os.makedirs(os.path.dirname(path))
    except Exception:
        pass
    r = requests.get(url, stream=True)
    with open(path, 'wb') as f:
        r.raw.decode_content = True
        shutil.copyfileobj(r.raw, f)


def getAccountDict(information_dict: Dict) -> Dict[str, str]:
    return {
        "name": information_dict["name"],
        "accessToken": information_dict["access_token"],
        "refreshToken": information_dict["refresh_token"],
        "uuid": information_dict["id"]
    }


def _addJavaRuntimeDir(path: str, runtimeList: List[str]) -> None:
    if not os.path.isdir(path):
        return
    for i in os.listdir(path):
        if not os.path.islink(os.path.join(path, i)):
            runtimeList.append(os.path.join(path, i, "bin", "java"))


def findJavaRuntimes() -> List[str]:
    runtimeList: list[str] = []
    _addJavaRuntimeDir("/usr/lib/jvm", runtimeList)
    _addJavaRuntimeDir("/usr/lib/sdk", runtimeList)
    _addJavaRuntimeDir("/app/jvm", runtimeList)
    return runtimeList


def isFlatpak() -> bool:
    return os.path.isfile("/.flatpak-info")


def isFrozen() -> bool:
    """Check if the App is frozen with PyInstaller"""
    return hasattr(sys, "frozen")


def clearTableWidget(table: QTableWidget) -> None:
    """Removes all Rows from a QTableWidget"""
    while table.rowCount() > 0:
        table.removeRow(0)


def stretchTableWidgetColumnsSize(table: QTableWidget) -> None:
    """Stretch all Columns of a QTableWidget"""
    for i in range(table.columnCount()):
        table.horizontalHeader().setSectionResizeMode(i, QHeaderView.ResizeMode.Stretch)


def selectComboBoxData(box: QComboBox, data: Any, default_index: int = 0) -> None:
    """Set the index to the item with the given data"""
    index = box.findData(data)
    if index == -1:
        box.setCurrentIndex(default_index)
    else:
        box.setCurrentIndex(index)


def isWayland() -> bool:
    """Returns if the Program is running in a Wayland session"""
    return QApplication.platformName() == "wayland"



from .Functions import isFlatpak, isWayland
from .Constants import DisplayServerSetting
from typing import List, TYPE_CHECKING
import minecraft_launcher_lib
import pathlib
import shutil
import shlex
import os


if TYPE_CHECKING:
    from .Environment import Environment
    from .Profile import Profile


def _getFlatpakSpawnCommand(profile: "Profile", env: "Environment") -> List[str]:
    command = [
        "flatpak-spawn",
        "--sandbox",
        "--sandbox-flag=share-display",
        "--sandbox-flag=share-sound",
        "--sandbox-flag=share-gpu",
        "--sandbox-flag=allow-dbus",
        "--sandbox-flag=allow-a11y"
    ]

    if profile.customGameDirectory:
        command += [
            f"--sandbox-expose-path-ro={env.minecraftDir}",
            f"--sandbox-expose-path={profile.getGameDirectoryPath()}",
            "--sandbox-expose-path=" + os.path.join(env.minecraftDir, "versions", profile.getVersionID()),
        ]
    else:
        command += [f"--sandbox-expose-path={env.minecraftDir}"]

    if profile.customExecutable:
        command += ["--sandbox-expose-path=" + str(pathlib.Path(profile.executable).parent.parent)]

    return command


def getMinecraftCommand(profile: "Profile", env: "Environment", natives_path: str) -> List[str]:
    version = profile.getVersionID()

    account = env.accountManager.getSelectedAccount()

    options: minecraft_launcher_lib.types.MinecraftOptions = {
        "username": account.getName(),
        "uuid": account.getMinecraftUUID(),
        "token": account.getAccessToken(),
        "launcherName": "jdMinecraftLauncher",
        "launcherVersion": env.launcherVersion,
        "gameDirectory": profile.getGameDirectoryPath(),
    }

    if profile.customExecutable:
        options["executablePath"] = profile.executable

    if profile.customArguments:
        options["jvmArguments"] = shlex.split(profile.arguments)

    if profile.customResolution:
        options["customResolution"] = True
        options["resolutionWidth"] = profile.resolutionX
        options["resolutionHeight"] = profile.resolutionY

    if profile.serverConnect:
        options["server"] = profile.serverIP
        if profile.serverPort != "":
            options["port"] = profile.serverPort

    if profile.demoMode:
        options["demo"] = True

    if profile.disableMultiplayer:
        options["disableMultiplayer"] = True

    if profile.disableChat:
        options["disableChat"] = True

    if natives_path != "":
        options["nativesDirectory"] = natives_path
        minecraft_launcher_lib.natives.extract_natives(version, env.minecraftDir, natives_path)

    command = minecraft_launcher_lib.command.get_minecraft_command(version, env.minecraftDir, options)

    if profile.hasMinecraftOptions:
        command += shlex.split(profile.minecraftOptions)

    if profile.useGameMode and shutil.which("gamemoderun"):
        command.insert(0, "gamemoderun")

    if isFlatpak() and env.settings.get("useFlatpakSubsandbox"):
        command = _getFlatpakSpawnCommand(profile, env) + command

    if isWayland():
        match env.settings.get("displayServer"):
            case DisplayServerSetting.WAYLAND:
                command = ["env", "-u", "DISPLAY"] + command
            case DisplayServerSetting.XWAYLAND:
                command = ["env", "-u", "WAYLAND_DISPLAY"] + command

    return command



from jdMinecraftLauncher.Functions import openFile, findJavaRuntimes, isFlatpak
from jdMinecraftLauncher.Shortcut import canCreateShortcuts, askCreateShortcut
from PyQt6.QtWidgets import QDialog, QFileDialog, QMessageBox
from ..ui_compiled.ProfileWindow import Ui_ProfileWindow
from jdMinecraftLauncher.Profile import Profile
from PyQt6.QtCore import QCoreApplication
from PyQt6.QtGui import QIntValidator
from typing import TYPE_CHECKING
import platform
import shutil
import os


if TYPE_CHECKING:
    from jdMinecraftLauncher.gui.MainWindow.MainWindow import MainWindow
    from jdMinecraftLauncher.Environment import Environment


class ProfileWindow(QDialog, Ui_ProfileWindow):
    def __init__(self, env: "Environment", parent: "MainWindow") -> None:
        super().__init__(parent)

        self.setupUi(self)

        self.env = env
        self.mainWindow = parent

        self.resolutionEditX.setValidator(QIntValidator(self.resolutionEditX))
        self.resolutionEditY.setValidator(QIntValidator(self.resolutionEditY))

        self.launcherVisibilityCombobox.addItem(QCoreApplication.translate("ProfileWindow", "Hide Launcher and re-open when game closes"))
        self.launcherVisibilityCombobox.addItem(QCoreApplication.translate("ProfileWindow", "Close Launcher when Game starts"))
        self.launcherVisibilityCombobox.addItem(QCoreApplication.translate("ProfileWindow", "Keep the Launcher open"))

        self.executableEdit.setEditable(True)
        self.executableEdit.addItems(findJavaRuntimes())

        self.gameDirectoryCheckbox.stateChanged.connect(lambda: self.gameDirectoryEdit.setEnabled(self.gameDirectoryCheckbox.isChecked()) or self.gameDirectoryBrowseButton.setEnabled(self.gameDirectoryCheckbox.isChecked()))
        self.gameDirectoryBrowseButton.clicked.connect(self.browseGameDirectoryClicked)
        self.resolutionCheckbox.stateChanged.connect(self.changeCustomResolution)
        self.launcherVisibilityCheckbox.stateChanged.connect(lambda: self.launcherVisibilityCombobox.setEnabled(self.launcherVisibilityCheckbox.isChecked()))
        self.enableSnapshots.stateChanged.connect(self.updateVersionsList)
        self.enableBeta.stateChanged.connect(self.updateVersionsList)
        self.enableAlpha.stateChanged.connect(self.updateVersionsList)
        self.cancelButton.clicked.connect(self.close)
        self.createShortcutButton.clicked.connect(lambda: askCreateShortcut(self.env, self, self.profile))  # type: ignore
        self.openGameDirectoryButton.clicked.connect(lambda: openFile(self.profile.getGameDirectoryPath()))  # type: ignore
        self.saveProfileButton.clicked.connect(self.saveProfile)
        self.executableCheckbox.stateChanged.connect(lambda: self.executableEdit.setEnabled(self.executableCheckbox.isChecked()) or self.executableButton.setEnabled(self.executableCheckbox.isChecked()))
        self.executableButton.clicked.connect(self.browseExecutableClicked)
        self.jvmArgumentsCheckbox.stateChanged.connect(lambda: self.jvmArgumentsEdit.setEnabled(self.jvmArgumentsCheckbox.isChecked()))
        self.serverCheckbox.stateChanged.connect(self.changeServerConnect)
        self.minecraftOptionsCheckBox.stateChanged.connect(lambda: self.minecraftOptionsEdit.setEnabled(self.minecraftOptionsCheckBox.isChecked()))

        self.selectLatestVersion = True

        if platform.system() != "Linux":
            self.gameModeCheckBox.setVisible(False)

        if isFlatpak():
            self.gameDirectoryEdit.setReadOnly(True)
            self.executableEdit.lineEdit().setReadOnly(True)

        if not shutil.which("gamemoderun"):
            self.gameModeCheckBox.setEnabled(False)

        self.resize(self.minimumSize())

    def changeCustomResolution(self) -> None:
        state = self.resolutionCheckbox.isChecked()
        self.resolutionLabel.setEnabled(state)
        self.resolutionEditX.setEnabled(state)
        self.resolutionEditY.setEnabled(state)

    def changeServerConnect(self) -> None:
        state = self.serverCheckbox.isChecked()
        self.serverLabel.setEnabled(state)
        self.serverEdit.setEnabled(state)
        self.portLabel.setEnabled(state)
        self.portEdit.setEnabled(state)

    def browseGameDirectoryClicked(self) -> None:
        path = QFileDialog.getExistingDirectory(directory=self.executableEdit.currentText())
        if path != "":
            self.gameDirectoryEdit.setText(path)

    def browseExecutableClicked(self) -> None:
        if isFlatpak():
            QMessageBox.information(self, QCoreApplication.translate("ProfileWindow", "Note for Flatpak users"), QCoreApplication.translate("ProfileWindow", "Please select in the following dialog the directory which contains bin/java"))
            path = QFileDialog.getExistingDirectory()
            if path == "":
                return
            javaPath = os.path.join(path, "bin", "java")
            if os.path.isfile(javaPath):
                self.executableEdit.lineEdit().setText(javaPath)
            else:
                QMessageBox.critical(self, QCoreApplication.translate("ProfileWindow", "Invalid directory"), QCoreApplication.translate("ProfileWindow", "This directory does not contain bin/java"))
        else:
            path = QFileDialog.getOpenFileName(directory=self.executableEdit.currentText())[0]
            if path != "":
                self.executableEdit.lineEdit().setText(path[0])

    def loadProfile(self, profile: "Profile", isNew: bool, copyText: bool = False) -> None:
        if isNew:
            if copyText:
                self.profileNameEdit.setText(QCoreApplication.translate("ProfileWindow", "Copy of {{name}}").replace("{{name}}", profile.name))
            else:
                self.profileNameEdit.setText(profile.name)
        else:
            self.profileNameEdit.setText(profile.name)

        self.gameDirectoryCheckbox.setChecked(profile.customGameDirectory)
        self.gameDirectoryEdit.setEnabled(profile.customGameDirectory)
        self.gameDirectoryBrowseButton.setEnabled(profile.customGameDirectory)
        self.gameDirectoryEdit.setText(profile.gameDirectoryPath)
        self.resolutionCheckbox.setChecked(profile.customResolution)
        self.resolutionEditX.setText(profile.resolutionX)
        self.resolutionEditY.setText(profile.resolutionY)
        self.launcherVisibilityCheckbox.setChecked(profile.customLauncherVisibility)
        self.launcherVisibilityCombobox.setEnabled(profile.customLauncherVisibility)
        self.launcherVisibilityCombobox.setCurrentIndex(profile.launcherVisibility)
        self.enableSnapshots.setChecked(profile.enableSnapshots)
        self.enableBeta.setChecked(profile.enableBeta)
        self.enableAlpha.setChecked(profile.enableAlpha)
        self.selectLatestVersion = profile.useLatestVersion
        self.selectLatestSnapshot = profile.useLatestSnapshot
        self.selectedVersion = profile.version
        self.executableCheckbox.setChecked(profile.customExecutable)
        self.executableEdit.setEnabled(profile.customExecutable)
        self.executableButton.setEnabled(profile.customExecutable)
        self.executableEdit.lineEdit().setText(profile.executable)
        self.jvmArgumentsCheckbox.setChecked(profile.customArguments)
        self.jvmArgumentsEdit.setEnabled(profile.customArguments)
        self.jvmArgumentsEdit.setText(profile.arguments)
        self.serverCheckbox.setChecked(profile.serverConnect)
        self.serverEdit.setText(profile.serverIP)
        self.portEdit.setText(profile.serverPort)
        self.demoModeCheckbox.setChecked(profile.demoMode)
        self.disableMultiplayerCheckBox.setChecked(profile.disableMultiplayer)
        self.disableChatCheckBox.setChecked(profile.disableChat)
        self.minecraftOptionsCheckBox.setChecked(profile.hasMinecraftOptions)
        self.minecraftOptionsEdit.setEnabled(profile.hasMinecraftOptions)
        self.minecraftOptionsEdit.setText(profile.minecraftOptions)
        self.gameModeCheckBox.setChecked(profile.useGameMode)
        self.changeCustomResolution()
        self.changeServerConnect()
        self.updateVersionsList()

        self.createShortcutButton.setVisible(not isNew and canCreateShortcuts())

        self.profile = profile
        self.isNew = isNew

    def saveProfile(self) -> None:
        profile = self.profile

        if self.isNew:
            profile = Profile(self.profileNameEdit.text(), self.env)
        else:
            profile.name = self.profileNameEdit.text()

        profile.customGameDirectory = self.gameDirectoryCheckbox.isChecked()
        profile.gameDirectoryPath = self.gameDirectoryEdit.text()
        profile.customResolution = self.resolutionCheckbox.isChecked()
        profile.resolutionX = self.resolutionEditX.text()
        profile.resolutionY = self.resolutionEditY.text()
        profile.customLauncherVisibility = self.launcherVisibilityCheckbox.isChecked()
        profile.launcherVisibility = self.launcherVisibilityCombobox.currentIndex()
        profile.enableSnapshots = self.enableSnapshots.isChecked()
        profile.enableBeta = self.enableBeta.isChecked()
        profile.enableAlpha = self.enableAlpha.isChecked()
        profile.customExecutable = self.executableCheckbox.isChecked()
        profile.executable = self.executableEdit.currentText()
        profile.customArguments = self.jvmArgumentsCheckbox.isChecked()
        profile.serverConnect = self.serverCheckbox.isChecked()
        profile.arguments = self.jvmArgumentsEdit.text()
        profile.serverIP = self.serverEdit.text()
        profile.serverPort = self.portEdit.text()
        profile.demoMode = self.demoModeCheckbox.isChecked()
        profile.disableMultiplayer = self.disableMultiplayerCheckBox.isChecked()
        profile.disableChat = self.disableChatCheckBox.isChecked()
        profile.hasMinecraftOptions = self.minecraftOptionsCheckBox.isChecked()
        profile.minecraftOptions = self.minecraftOptionsEdit.text().strip()
        profile.useGameMode = self.gameModeCheckBox.isChecked()
        version = self.versionSelectCombobox.currentData()

        if version == "latestRelease":
            profile.useLatestVersion = True
            profile.useLatestSnapshot = False
        elif version == "latestSnapshot":
            profile.useLatestSnapshot = True
            profile.useLatestVersion = False
        else:
            profile.useLatestVersion = False
            profile.useLatestSnapshot = False
            profile.version = version

        if self.isNew:
            self.env.profileCollection.profileList.append(profile)
            self.env.profileCollection.selectedProfile = profile.id

        self.mainWindow.updateProfileList()
        self.env.profileCollection.save()
        self.close()

    def updateVersionsList(self) -> None:
        self.versionSelectCombobox.clear()
        self.versionSelectCombobox.addItem(QCoreApplication.translate("ProfileWindow", "Use latest Version"), "latestRelease")

        if self.enableSnapshots.isChecked():
            self.versionSelectCombobox.addItem(QCoreApplication.translate("ProfileWindow", "Use latest Snapshot"), "latestSnapshot")

        for i in self.env.versions["versions"]:
            if i["type"] == "release":
                self.versionSelectCombobox.addItem("release " + i["id"], i["id"])
            elif i["type"] == "snapshot" and self.enableSnapshots.isChecked():
                self.versionSelectCombobox.addItem("snapshot " + i["id"], i["id"])
            elif i["type"] == "old_beta" and self.enableBeta.isChecked():
                self.versionSelectCombobox.addItem("old_beta " + i["id"], i["id"])
            elif i["type"] == "old_alpha" and self.enableAlpha.isChecked():
                self.versionSelectCombobox.addItem("old_alpha " + i["id"], i["id"])

        if self.selectLatestVersion:
            self.versionSelectCombobox.setCurrentIndex(0)
        elif self.selectLatestSnapshot and self.enableSnapshots.isChecked():
            self.versionSelectCombobox.setCurrentIndex(1)
        else:
            for i in range(self.versionSelectCombobox.count()):
                if self.versionSelectCombobox.itemText(i).split(" ", 1)[1] == self.selectedVersion:
                    self.versionSelectCombobox.setCurrentIndex(i)



from jdMinecraftLauncher.RunMinecraft import getMinecraftCommand
from PyQt6.QtCore import pyqtClassInfo, pyqtSlot, pyqtProperty  # type:ignore[attr-defined]
from PyQt6.QtDBus import QDBusConnection, QDBusAbstractAdaptor
from PyQt6.QtWidgets import QApplication
from typing import TYPE_CHECKING
import json
import sys
import os


if TYPE_CHECKING:
    from jdMinecraftLauncher.Environment import Environment
    from .gui.MainWindow.MainWindow import MainWindow


with open(os.path.join(os.path.dirname(__file__), "DBusInterface.xml"), "r", encoding="utf-8") as f:
    interface = f.read()


@pyqtClassInfo("D-Bus Interface", "page.codeberg.JakobDev.jdMinecraftLauncher")  # type: ignore
@pyqtClassInfo("D-Bus Introspection", interface)  # type: ignore
class DBusService(QDBusAbstractAdaptor):
    def __init__(self, env: "Environment", app: QApplication, mainWindow: "MainWindow") -> None:
        super().__init__(app)
        QDBusConnection.sessionBus().registerObject("/", app)

        if not QDBusConnection.sessionBus().registerService("page.codeberg.JakobDev.jdMinecraftLauncher"):
            print(QDBusConnection.sessionBus().lastError().message(), file=sys.stderr)

        self._env = env
        self._mainWindow = mainWindow

    @pyqtSlot(str, result=bool)  # type: ignore
    def LaunchProfile(self, name: str) -> bool:
        profile = self._env.profileCollection.getProfileByName(name)
        if profile:
            self._mainWindow.launchProfile(profile)
            return True
        else:
            return False

    @pyqtSlot(result=list)  # type: ignore
    def ListProfiles(self) -> list[str]:
        profileList = []
        for i in self._env.profiles:
            profileList.append(i.name)
        return profileList

    @pyqtSlot(result=str)  # type: ignore
    def GetProfile(self) -> str:
        return json.dumps(self._env.profileCollection.getSelectedProfile().toDict())

    @pyqtSlot(str, result=list)  # type: ignore
    def GetMinecraftCommand(self, profile_id: str) -> list[str]:
        profile = self._env.profileCollection.getProfileByID(profile_id)
        if profile:
            return getMinecraftCommand(profile, self._env, "")
        return []

    @pyqtSlot(result=str)  # type: ignore
    def GetVersion(self) -> str:
        return self._env.launcherVersion

    @pyqtProperty(bool)  # type: ignore
    def MainWindowVisible(self) -> bool:
        return self._mainWindow.isVisible()

    @MainWindowVisible.setter  # type: ignore
    def MainWindowVisible(self, show: bool) -> None:
        self._mainWindow.setVisible(show)



from ...Functions import isFlatpak, selectComboBoxData, isWayland
from ...Constants import DisplayServerSetting, NewsTypeSetting
from PyQt6.QtWidgets import QWidget, QMessageBox, QFileDialog
from ...ui_compiled.OptionsTab import Ui_OptionsTab
from PyQt6.QtCore import Qt, QCoreApplication
from ...Languages import getLanguageNames
from typing import TYPE_CHECKING
import minecraft_launcher_lib
import os


if TYPE_CHECKING:
    from ...Environment import Environment
    from .MainWindow import MainWindow


class OptionsTab(QWidget, Ui_OptionsTab):
    def __init__(self, env: "Environment", parent: "MainWindow") -> None:
        super().__init__()

        self.setupUi(self)

        self._env = env
        self._parent = parent

        languageNames = getLanguageNames()
        self.languageComboBox.addItem(languageNames.get("en", "en"), "en")
        for i in os.listdir(os.path.join(env.currentDir, "translations")):
            if not i.endswith(".qm"):
                continue

            lang = i.removeprefix("jdMinecraftLauncher_").removesuffix(".qm")
            self.languageComboBox.addItem(languageNames.get(lang, lang), lang)
        self.languageComboBox.model().sort(0, Qt.SortOrder.AscendingOrder)
        self.languageComboBox.insertItem(0, QCoreApplication.translate("OptionsTab", "Use System Language"), "default")

        self.displayServerBox.addItem(QCoreApplication.translate("OptionsTab", "Auto"), DisplayServerSetting.AUTO)
        self.displayServerBox.addItem(QCoreApplication.translate("OptionsTab", "Wayland"), DisplayServerSetting.WAYLAND)
        self.displayServerBox.addItem(QCoreApplication.translate("OptionsTab", "XWayland"), DisplayServerSetting.XWAYLAND)

        selectComboBoxData(self.languageComboBox, env.settings.get("language"))

        match env.settings.get("newsType"):
            case NewsTypeSetting.MINECRAFT:
                self.newsMinecraftButton.setChecked(True)
            case NewsTypeSetting.RSS:
                self.newsRssFeedButton.setChecked(True)
            case NewsTypeSetting.WEBSITE:
                self.newsWebsiteButton.setChecked(True)

        self._updateNewsEnabled()
        self.newsFeedUrlEdit.setText(env.settings.get("newsFeedURL"))
        self.newsFeedDefaultBrowserCheckBox.setChecked(env.settings.get("newsFeedDefaultBrowser"))
        self.newsUrlEdit.setText(env.settings.get("newsURL"))
        self.allowMultiLaunchCheckBox.setChecked(self._env.settings.get("enableMultiLaunch"))
        self.extractNativesCheckBox.setChecked(self._env.settings.get("extractNatives"))
        self.windowIconProgressCheckBox.setChecked(self._env.settings.get("windowIconProgress"))
        self.flatpakSubsandboxCheckBox.setChecked(self._env.settings.get("useFlatpakSubsandbox"))
        self.checkUpdatesStartupCheckBox.setChecked(self._env.settings.get("checkUpdatesStartup"))
        selectComboBoxData(self.displayServerBox, self._env.settings.get("displayServer"))

        self.windowIconProgressCheckBox.setVisible(parent.windowIconProgress.isSupported())

        if not isFlatpak():
            self.flatpakSubsandboxCheckBox.setVisible(False)

        self.checkUpdatesStartupCheckBox.setVisible(env.enableUpdater)
        self.displayServerLabel.setVisible(isWayland())
        self.displayServerBox.setVisible(isWayland())

        self.newsMinecraftButton.toggled.connect(self._newsTypeChanged)
        self.newsRssFeedButton.toggled.connect(self._newsTypeChanged)
        self.newsFeedUrlEdit.editingFinished.connect(self._newsFeedUrlChanged)
        self.newsFeedDefaultBrowserCheckBox.stateChanged.connect(self._newsFeedDefaultBrowserCheckBoxChanged)
        self.newsUrlEdit.editingFinished.connect(self._newsUrlChanged)
        self.allowMultiLaunchCheckBox.stateChanged.connect(self._multiLaunchCheckBoxChanged)
        self.extractNativesCheckBox.stateChanged.connect(self._extractNativesCheckBoxChanged)
        self.windowIconProgressCheckBox.stateChanged.connect(self._windowIconProgressCheckBoxChanged)
        self.flatpakSubsandboxCheckBox.stateChanged.connect(self._flatpakSubsandboxCheckBoxChanged)
        self.minecraftDirChangeButton.clicked.connect(self._minecraftDirChangeButtonClicked)
        self.minecraftDirResetButton.clicked.connect(self._minecraftDirResetButtonClicked)
        self.displayServerBox.currentIndexChanged.connect(self._displayServerBoxChanged)

        self._updateMinecraftDirWidgets()

    def _getSelectedNewsType(self) -> str:
        if self.newsMinecraftButton.isChecked():
            return NewsTypeSetting.MINECRAFT
        elif self.newsRssFeedButton.isChecked():
            return NewsTypeSetting.RSS
        elif self.newsWebsiteButton.isChecked():
            return NewsTypeSetting.WEBSITE
        else:
            raise Exception("unknown news type")

    def _updateNewsEnabled(self) -> None:
        newsType = self._getSelectedNewsType()

        self.newsFeedUrlLabel.setEnabled(newsType == NewsTypeSetting.RSS)
        self.newsFeedUrlEdit.setEnabled(newsType == NewsTypeSetting.RSS)
        self.newsUrlLabel.setEnabled(newsType == NewsTypeSetting.WEBSITE)
        self.newsUrlEdit.setEnabled(newsType == NewsTypeSetting.WEBSITE)

    def _newsTypeChanged(self) -> None:
        self._env.settings.set("newsType", self._getSelectedNewsType())

        self._updateNewsEnabled()
        self._parent.updateNewsTab()

    def _newsFeedUrlChanged(self) -> None:
        self._env.settings.set("newsFeedURL", self.newsFeedUrlEdit.text())
        self._parent.updateNewsTab()

    def _newsFeedDefaultBrowserCheckBoxChanged(self) -> None:
        self._env.settings.set("newsFeedDefaultBrowser", self.newsFeedDefaultBrowserCheckBox.isChecked())

    def _newsUrlChanged(self) -> None:
        self._env.settings.set("newsURL", self.newsUrlEdit.text())
        self._parent.updateNewsTab()

    def _multiLaunchCheckBoxChanged(self) -> None:
        self._env.settings.set("enableMultiLaunch", self.allowMultiLaunchCheckBox.isChecked())

    def _extractNativesCheckBoxChanged(self) -> None:
        self._env.settings.set("extractNatives", self.extractNativesCheckBox.isChecked())

    def _windowIconProgressCheckBoxChanged(self) -> None:
        checked = self.windowIconProgressCheckBox.isChecked()
        self._env.settings.set("windowIconProgress", checked)
        if not checked:
            self._parent.windowIconProgress.hide()

    def _flatpakSubsandboxCheckBoxChanged(self) -> None:
        self._env.settings.set("useFlatpakSubsandbox", self.flatpakSubsandboxCheckBox.isChecked())

    def _displayServerBoxChanged(self) -> None:
        self._env.settings.set("displayServer", self.displayServerBox.currentData())

    def _updateMinecraftDirWidgets(self) -> None:
        if (customMinecraftDir := self._env.settings.get("customMinecraftDir")) is not None:
            self.minecraftDirPathButton.setText(customMinecraftDir)
        else:
            self.minecraftDirPathButton.setText(minecraft_launcher_lib.utils.get_minecraft_directory())

        self.minecraftDirResetButton.setEnabled(self._env.settings.get("customMinecraftDir") is not None)

    def _minecraftDirChangeButtonClicked(self) -> None:
        text = QCoreApplication.translate("OptionsTab", "This will change your Minecraft directory.") + "<br><br>"
        text += QCoreApplication.translate("OptionsTab", "This is the directory where Minecraft is installed.") + "<br><br>"
        text += QCoreApplication.translate("OptionsTab", "If you don't set a game directory for your profile, Minecraft will also save its data (saves, options, etc.) there.") + "<br><br>"
        text += QCoreApplication.translate("OptionsTab", "Would you like to proceed?")

        if QMessageBox.question(self, QCoreApplication.translate("OptionsTab", "Change Minecraft directory"), text, QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No) != QMessageBox.StandardButton.Yes:
            return

        path = QFileDialog.getExistingDirectory(self)

        if path == "":
            return

        self._env.minecraftDir = path

        self._env.updateInstalledVersions()
        self._parent.updateProfileList()

        self._env.settings.set("customMinecraftDir", path)

        if not self._env.args.dont_save_data:
            self._env.settings.save(os.path.join(self._env.dataDir, "settings.json"))

        self._updateMinecraftDirWidgets()

    def _minecraftDirResetButtonClicked(self) -> None:
        if QMessageBox.question(self, QCoreApplication.translate("OptionsTab", "Reset Minecraft directory"), QCoreApplication.translate("OptionsTab", "Would you like to reset your Minecraft directory to its default location?"), QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No) != QMessageBox.StandardButton.Yes:
            return

        self._env.minecraftDir = minecraft_launcher_lib.utils.get_minecraft_directory()

        self._env.updateInstalledVersions()
        self._parent.updateProfileList()

        self._env.settings.set("customMinecraftDir", None)

        if not self._env.args.dont_save_data:
            self._env.settings.save(os.path.join(self._env.dataDir, "settings.json"))

        self._updateMinecraftDirWidgets()

    def saveSettings(self) -> None:
        self._env.settings.set("language", self.languageComboBox.currentData())

        if self.newsMinecraftButton.isChecked():
            self._env.settings.set("newsType", NewsTypeSetting.MINECRAFT)
        elif self.newsRssFeedButton.isChecked():
            self._env.settings.set("newsType", NewsTypeSetting.RSS)
        elif self.newsWebsiteButton.isChecked():
            self._env.settings.set("newsType", NewsTypeSetting.WEBSITE)

        self._env.settings.set("newsFeedURL", self.newsFeedUrlEdit.text())
        self._env.settings.set("newsURL", self.newsUrlEdit.text())
        self._env.settings.set("newsFeedDefaultBrowser", self.newsFeedDefaultBrowserCheckBox.isChecked())
        self._env.settings.set("enableMultiLaunch", self.allowMultiLaunchCheckBox.isChecked())
        self._env.settings.set("extractNatives", self.extractNativesCheckBox.isChecked())
        self._env.settings.set("useFlatpakSubsandbox", self.flatpakSubsandboxCheckBox.isChecked())
        self._env.settings.set("checkUpdatesStartup", self.checkUpdatesStartupCheckBox.isChecked())
        self._env.settings.set("displayServer", self.displayServerBox.currentData())
        self._env.settings.save(os.path.join(self._env.dataDir, "settings.json"))



from PyQt6.QtWidgets import QTableWidget, QTableWidgetItem, QHeaderView, QAbstractItemView, QPushButton
from PyQt6.QtCore import Qt, QCoreApplication
from typing import TYPE_CHECKING
import minecraft_launcher_lib
import sys


if TYPE_CHECKING:
    from ...Environment import Environment
    from .MainWindow import MainWindow


class ForgeTab(QTableWidget):
    def __init__(self, env: "Environment", mainWindow: "MainWindow") -> None:
        super().__init__(0, 2)

        self._env = env
        self._mainWindow = mainWindow

        self.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
        self.horizontalHeader().setSectionResizeMode(1, QHeaderView.ResizeMode.Stretch)
        self.setSelectionMode(QAbstractItemView.SelectionMode.SingleSelection)
        self.horizontalHeader().hide()
        self.verticalHeader().hide()

        count = 0
        minecraft_version_check = {}

        try:
            forgeVersionList = minecraft_launcher_lib.forge.list_forge_versions()
        except Exception:
            print("Could not get Forge Versions", file=sys.stderr)
            return

        for i in forgeVersionList:
            minecraft_version, _ = i.split("-", 1)

            if minecraft_version in minecraft_version_check or not minecraft_launcher_lib.forge.supports_automatic_install(i):
                continue

            minecraft_version_check[minecraft_version] = True

            versionItem = QTableWidgetItem(i)
            versionItem.setFlags(versionItem.flags() ^ Qt.ItemFlag.ItemIsEditable)

            installButton = QPushButton(QCoreApplication.translate("ForgeTab", "Install"))

            installButton.clicked.connect(self._installButtonClicked)

            self.insertRow(count)
            self.setItem(count, 0, versionItem)
            self.setCellWidget(count, 1, installButton)

            count += 1

    def _installForgeVersion(self, forgeVersion: str) -> None:
        self._mainWindow.installThread.setupForgeInstallation(forgeVersion)
        self._mainWindow.installThread.start()

        self._mainWindow.setInstallButtonsEnabled(False)

    def _installButtonClicked(self) -> None:
        for i in range(self.rowCount()):
            if self.cellWidget(i, 1) == self.sender():
                self._installForgeVersion(self.item(i, 0).text())
                return

    def setButtonsEnabled(self, enabled: bool) -> None:
        for i in range(self.rowCount()):
            self.cellWidget(i, 1).setEnabled(enabled)



from typing import Type, TYPE_CHECKING
import minecraft_launcher_lib
import uuid


if TYPE_CHECKING:
    from jdMinecraftLauncher.Environment import Environment


class Profile:
    def __init__(self, name: str, env: "Environment") -> None:
        self.env: "Environment" = env

        self.id = self._generateProfileID()
        self.name = name
        self.version = ""
        self.useLatestVersion = True
        self.useLatestSnapshot = False
        self.customGameDirectory = False
        self.gameDirectoryPath = env.minecraftDir
        self.customResolution = False
        self.resolutionX = "854"
        self.resolutionY = "480"
        self.customLauncherVisibility = False
        self.launcherVisibility = 0
        self.enableSnapshots = False
        self.enableBeta = False
        self.enableAlpha = False
        self.customExecutable = False
        self.executable = minecraft_launcher_lib.utils.get_java_executable()
        self.customArguments = False
        self.arguments = "-Xms512M -Xmx512M"
        self.serverConnect = False
        self.serverIP = ""
        self.serverPort = ""
        self.demoMode = False
        self.disableMultiplayer = False
        self.disableChat = False
        self.hasMinecraftOptions = False
        self.minecraftOptions = ""
        self.useGameMode = False

    def _generateProfileID(self) -> str:
        while True:
            current_id = str(uuid.uuid4())
            for i in self.env.profileCollection.profileList:
                if i.id == current_id:
                    break
            else:
                return current_id

    def getVersionID(self) -> str:
        if self.useLatestVersion:
            return self.env.versions["latest"]["release"]
        elif self.useLatestSnapshot:
            return self.env.versions["latest"]["snapshot"]
        else:
            return self.version

    def getGameDirectoryPath(self) -> str:
        if self.customGameDirectory:
            return self.gameDirectoryPath
        else:
            return self.env.minecraftDir

    def getJavaPath(self) -> str:
        if self.customExecutable:
            return self.executable
        else:
            return "java"

    @classmethod
    def load(cls: Type["Profile"], env: "Environment", objects: dict, profileVersion: int) -> "Profile":
        profile = Profile(objects["name"], env)

        if "id" in objects:
            profile.id = objects["id"]
        else:
            profile.id = profile._generateProfileID()

        profile.version = objects["version"]
        profile.useLatestVersion = objects["useLatestVersion"]
        profile.customGameDirectory = objects["customGameDirectory"]
        profile.gameDirectoryPath = objects["gameDirectoryPath"]
        profile.customResolution = objects["customResolution"]
        profile.resolutionX = objects["resolutionX"]
        profile.resolutionY = objects["resolutionY"]
        profile.customLauncherVisibility = objects["customLauncherVisibility"]
        profile.launcherVisibility = objects["launcherVisibility"]
        profile.enableSnapshots = objects["enableSnapshots"]
        profile.enableBeta = objects["enableBeta"]
        profile.enableAlpha = objects["enableAlpha"]
        profile.customExecutable = objects["customExecutable"]
        profile.executable = objects["executable"]
        profile.customArguments = objects["customArguments"]
        profile.arguments = objects["arguments"]
        profile.serverConnect = objects["serverConnect"]
        profile.serverIP = objects["serverIP"]
        profile.serverPort = objects["serverPort"]
        profile.demoMode = objects["demoMode"]
        profile.disableMultiplayer = objects.get("disableMultiplayer", False)
        profile.disableChat = objects.get("disableMultiplayer", False)
        profile.hasMinecraftOptions = objects.get("hasMinecraftOptions", False)
        profile.minecraftOptions = objects.get("minecraftOptions", "")
        profile.useGameMode = objects.get("useGameMode", False)

        if profileVersion == 1:
            try:
                profile.version = profile.version.split(" ")[1]
            except IndexError:
                pass

        return profile

    def toDict(self) -> dict:
        data = {}
        for key, value in vars(self).items():
            if isinstance(value, int) or isinstance(value, bool) or isinstance(value, str):
                data[key] = value
        return data



from PyQt6.QtWebEngineWidgets import QWebEngineView
from ...utils.NewsGenerator import GenerateNewsPage
from PyQt6.QtWebEngineCore import QWebEnginePage
from PyQt6.QtCore import QUrl, QCoreApplication
from ...Constants import NewsTypeSetting
from typing import cast, TYPE_CHECKING
import webbrowser
import traceback
import sys


if TYPE_CHECKING:
    from ...Environment import Environment


class _NewsPage(QWebEnginePage):
    def __init__(self, env: "Environment") -> None:
        super().__init__()

        self._env = env

    def _renderNewsPage(self) -> None:
        try:
            self.setHtml(GenerateNewsPage(self._env), QUrl("localhost"))
        except ModuleNotFoundError as ex:
            errorText = QCoreApplication.translate("NewsTab", "{{name}} not installed").replace("{{name}}", cast(str, ex.name)) + "<br>"
            errorText += QCoreApplication.translate("NewsTab", "You need the {{name}} Python package installed to use this feature").replace("{{name}}", cast(str, ex.name))
            self.setHtml(errorText, QUrl("localhost"))
        except Exception:
            errorText = "<h1>" + QCoreApplication.translate("NewsTab", "Error") + "</h1>"
            match self._env.settings.get("newsType"):
                case NewsTypeSetting.MINECRAFT:
                    errorText += "<p>" + QCoreApplication.translate("NewsTab", "Unable to render Minecraft News") + "<p>"
                case NewsTypeSetting.RSS:
                    url = self._env.settings.get("newsFeedURL")
                    link = f'<a href="{url}">{url}</a>'
                    errorText += "<p>" + QCoreApplication.translate("NewsTab", "Unable to render RSS feed {{url}}").replace("{{url}}", link) + "<p>"
            self.setHtml(errorText, QUrl("localhost"))
            print(traceback.format_exc(), file=sys.stderr)

    def updateNews(self) -> None:
        match self._env.settings.get("newsType"):
            case NewsTypeSetting.MINECRAFT | NewsTypeSetting.RSS:
                self._renderNewsPage()
            case NewsTypeSetting.WEBSITE:
                self.setUrl(QUrl(self._env.settings.get("newsURL")))

    def acceptNavigationRequest(self, url: QUrl, navigationType: QWebEnginePage.NavigationType, isMainFrame: bool) -> bool:
        if navigationType == QWebEnginePage.NavigationType.NavigationTypeLinkClicked:
            if self._env.settings.get("newsFeedDefaultBrowser"):
                webbrowser.open(url.toString())
                return False
            else:
                return True
        else:
            return True


class NewsTab(QWebEngineView):
    def __init__(self, env: "Environment") -> None:
        super().__init__()

        self._env = env

        self._page = _NewsPage(env)
        self.setPage(self._page)

    def updateNews(self) -> None:
        self._page.updateNews()



from jdMinecraftLauncher.ProfileCollection import ProfileCollection
from jdMinecraftLauncher.Settings import Settings
from jdMinecraftLauncher.Profile import Profile
from .AccountManager import AccountManager
from PyQt6.QtWidgets import QApplication
from .Functions import isFlatpak
from PyQt6.QtCore import QLocale
from PyQt6.QtGui import QIcon
import minecraft_launcher_lib
from pathlib import Path
from typing import cast
import argparse
import platform
import requests
import tomllib
import json
import os


class Environment:
    def __init__(self, app: QApplication) -> None:
        self.offlineMode = False
        self.currentDir = os.path.dirname(os.path.realpath(__file__))

        with open(os.path.join(self.currentDir, "version.txt"), "r", encoding="utf-8") as f:
            self.launcherVersion = f.read().strip()

        self.icon = QIcon(os.path.join(self.currentDir, "Icon.svg"))
        self.app = app

        parser = argparse.ArgumentParser()
        parser.add_argument("url", nargs="?")
        parser.add_argument("--minecraft-dir", help="Set the Minecraft Directory")
        parser.add_argument("--data-dir", help="Set the Data Directory")
        parser.add_argument("--launch-profile", help="Launch a Profile")
        parser.add_argument("--account", help="Launch with the selected Account")
        parser.add_argument("--offline-mode", help="Force offline Mode", action="store_true")
        parser.add_argument("--force-start", help="Forces the start on unsupported Platforms", action="store_true")
        parser.add_argument("--dont-save-data", help="Don't save data to the disk (only for development usage)", action="store_true")
        parser.add_argument("--debug", help="Start in Debug Mode", action="store_true")
        self.args = parser.parse_known_args()[0]

        if self.args.data_dir:
            self.dataDir = self.args.data_dir
        else:
            self.dataDir = self.getDataPath()

        if not os.path.exists(self.dataDir):
            os.makedirs(self.dataDir)

        self.debugMode = bool(self.args.debug)

        self.settings = Settings()
        self.settings.load(os.path.join(self.dataDir, "settings.json"))

        if self.args.minecraft_dir:
            self.minecraftDir = self.args.minecraft_dir
        elif self.settings.get("customMinecraftDir") is not None:
            self.minecraftDir = self.settings.get("customMinecraftDir")
        else:
            self.minecraftDir = minecraft_launcher_lib.utils.get_minecraft_directory()

        self.accountManager = AccountManager(self)

        self.profileCollection = ProfileCollection(self)
        self.profileCollection.loadProfiles()

        self.profiles: list[Profile] = []
        self.selectedProfile = 0

        with open(os.path.join(self.currentDir, "Distribution.toml"), "rb") as f:
            distributionConfig = tomllib.load(f)

        self.enableUpdater = not isFlatpak() and distributionConfig.get("EnableUpdater", True)

        if self.settings.get("language") == "default":
            self.locale = QLocale()
        else:
            self.locale = QLocale(self.settings.get("language"))

        if os.path.isdir(self.dataDir):
            self.firstLaunch = False
        else:
            self.firstLaunch = True

    def _getSystemDataDir(self) -> str:
        match platform.system():
            case "Windows":
                return cast(str, os.getenv("APPDATA"))
            case "Darwin":
                return os.path.join(Path.home(), "Library", "Application Support")
            case "Haiku":
                return os.path.join(Path.home(), "config", "settings")
            case _:
                if os.getenv("XDG_DATA_HOME"):
                    return cast(str, os.getenv("XDG_DATA_HOME"))
                else:
                    return os.path.join(Path.home(), ".local", "share")

    def getDataPath(self) -> str:
        if os.path.isdir(os.path.join(minecraft_launcher_lib.utils.get_minecraft_directory(), "jdMinecraftLauncher")):
            return os.path.join(minecraft_launcher_lib.utils.get_minecraft_directory(), "jdMinecraftLauncher")

        dataPath = self._getSystemDataDir()
        if os.path.isdir(os.path.join(dataPath, "jdMinecraftLauncher")):
            return os.path.join(dataPath, "jdMinecraftLauncher")

        return os.path.join(dataPath, "JakobDev", "jdMinecraftLauncher")

    def loadVersions(self) -> None:
        if not self.offlineMode:
            try:
                r = requests.get("https://launchermeta.mojang.com/mc/game/version_manifest_v2.json")
                f = open(os.path.join(self.dataDir, "versions_cache.json"), "w")
                f.write(r.text)
                f.close()
                r.close()
            except Exception:
                print("Failed to update versions list")
        with open(os.path.join(self.dataDir, "versions_cache.json")) as f:
            self.versions = json.load(f)
        self.updateInstalledVersions()

    def updateInstalledVersions(self) -> None:
        versioncheck = {}
        for v in self.versions["versions"]:
            versioncheck[v["id"]] = True

        self.installedVersion = []
        if os.path.isdir(os.path.join(self.minecraftDir, "versions")):
            for v in os.listdir(os.path.join(self.minecraftDir, "versions")):
                json_path = os.path.join(self.minecraftDir, "versions", v, v + ".json")
                if os.path.isfile(json_path):
                    with open(json_path) as f:
                        try:
                            vinfo = json.load(f)
                        except json.decoder.JSONDecodeError as e:
                            print("Error while parsing " + json_path + ": " + e.args[0])
                            continue

                    tmp = {"id": vinfo["id"], "type": vinfo["type"]}
                    self.installedVersion.append(tmp)

                    if vinfo["id"] not in versioncheck:
                        self.versions["versions"].append(tmp)



from PyQt6.QtCore import QCoreApplication, QTranslator, QLibraryInfo
from PyQt6.QtWidgets import QApplication, QSplashScreen, QMessageBox
from jdMinecraftLauncher.Functions import hasInternetConnection
from jdMinecraftLauncher.Environment import Environment
from .MicrosoftSecrets import MicrosoftSecrets
from .utils.UpdateChecker import checkUpdates
from .ProfileImporter import askProfileImport
import minecraft_launcher_lib
import traceback
import sys
import os


def _ensureMinecraftDirectoryExists(env: Environment) -> None:
    if os.path.isdir(env.minecraftDir):
        return

    try:
        os.makedirs(env.minecraftDir)
    except Exception:
        print(traceback.format_exc(), file=sys.stderr)

        text = QCoreApplication.translate("jdMinecraftLauncher", "The Minecraft directory was not found and could not be created.")

        if env.settings.get("customMinecraftDir") is None:
            QMessageBox.critical(None, QCoreApplication.translate("jdMinecraftLauncher", "Minecraft directory not found"), text)
            sys.exit(1)
        else:
            text += "<br><br>" + QCoreApplication.translate("jdMinecraftLauncher", "You have set a custom Minecraft directory. Would you like to revert it back to the default?")

            if QMessageBox.question(None, QCoreApplication.translate("jdMinecraftLauncher", "Minecraft directory not found"), text, QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No) != QMessageBox.StandardButton.Yes:
                sys.exit(1)

            env.minecraftDir = minecraft_launcher_lib.utils.get_minecraft_directory()

            env.settings.set("customMinecraftDir", None)
            env.settings.save(os.path.join(env.dataDir, "settings.json"))

            _ensureMinecraftDirectoryExists(env)


def _handleAccount(env: Environment, splashScreen: QSplashScreen) -> bool:
    account = env.accountManager.getSelectedAccount()

    if account is None:
        splashScreen.close()
        if env.offlineMode:
            QMessageBox.critical(None, QCoreApplication.translate("jdMinecraftLauncher", "No Internet Connection"), QCoreApplication.translate("jdMinecraftLauncher", "You have no Internet connection. If you start jdMinecraftLauncher for the first time, you have to login using the Internet before you can use the offline Mode."))
            return False

        return env.accountManager.addMicrosoftAccount(None) is not None

    if env.offlineMode:
        splashScreen.close()
        return True

    if account.reload():
        splashScreen.close()
        return True

    splashScreen.close()

    return account.login(None)


def main() -> None:
    if not os.path.isdir(os.path.join(os.path.dirname(__file__), "ui_compiled")):
        print("Could not find compiled ui files. Please run tools/CompileUI.py first.", file=sys.stderr)
        sys.exit(1)

    app = QApplication(sys.argv)
    env = Environment(app)

    app.setWindowIcon(env.icon)
    app.setOrganizationName("JakobDev")
    app.setApplicationName("jdMinecraftLauncher")
    app.setDesktopFileName("page.codeberg.JakobDev.jdMinecraftLauncher")

    MicrosoftSecrets.setup(env)

    qtTranslator = QTranslator()
    if qtTranslator.load(env.locale, "qt", "_", QLibraryInfo.path(QLibraryInfo.LibraryPath.TranslationsPath)):
        app.installTranslator(qtTranslator)

    webengineTranslator = QTranslator()
    if webengineTranslator.load(env.locale, "qtwebengine", "_", QLibraryInfo.path(QLibraryInfo.LibraryPath.TranslationsPath)):
        app.installTranslator(webengineTranslator)

    appTranslator = QTranslator()
    if appTranslator.load(env.locale, "jdMinecraftLauncher", "_", os.path.join(env.currentDir, "translations")):
        app.installTranslator(appTranslator)

    if not minecraft_launcher_lib.utils.is_platform_supported() and not env.args.force_start:
        QMessageBox.critical(None, QCoreApplication.translate("jdMinecraftLauncher", "Unsupported Platform"), QCoreApplication.translate("jdMinecraftLauncher", "Your current Platform is not supported by jdMinecraftLauncher"))
        sys.exit(0)

    if env.args.offline_mode or not hasInternetConnection():
        env.offlineMode = True

    if env.firstLaunch:
        welcomeText = QCoreApplication.translate("jdMinecraftLauncher", "It appears to be your first time using jdMinecraftLauncher.") + " "
        welcomeText += QCoreApplication.translate("jdMinecraftLauncher", "This is a custom Minecraft Launcher designed to resemble the old official launcher in appearance and feel, but with modern features like Microsoft account support.") + "<br><br>"
        welcomeText += QCoreApplication.translate("jdMinecraftLauncher", "If you encounter any issues, please report them so that they can be addressed and resolved.") + "<br><br>"
        welcomeText += QCoreApplication.translate("jdMinecraftLauncher", "This launcher is not an official product of Mojang/Microsoft.")
        QMessageBox.information(None, QCoreApplication.translate("jdMinecraftLauncher", "Welcome"), welcomeText)

    if env.enableUpdater and not env.offlineMode and env.settings.get("checkUpdatesStartup"):
        checkUpdates(env)

    _ensureMinecraftDirectoryExists(env)

    splashScreen = QSplashScreen()
    splashScreen.setPixmap(env.icon.pixmap(128, 128))
    splashScreen.show()

    try:
        import setproctitle
        setproctitle.setproctitle("jdMinecraftLauncher")
    except ModuleNotFoundError:
        pass

    env.loadVersions()

    # We import it here, so jdMinecraftLauncher doen't crash if the compiled ui files and missing and is able to show the error message
    from jdMinecraftLauncher.gui.MainWindow.MainWindow import MainWindow

    mainWindow = MainWindow(env)

    if not _handleAccount(env, splashScreen):
        sys.exit(0)

    env.accountManager.saveData()

    if env.args.account:
        account = env.accountManager.getAccountByName(env.args.account)
        if account is not None:
            env.accountManager.setSelectedAccount(account)
        else:
            print(QCoreApplication.translate("jdMinecraftLauncher", "Account {{name}} does not exist").replace("{{name}}", env.args.account), file=sys.stderr)

    if env.firstLaunch:
        askProfileImport(env, mainWindow)

    mainWindow.updateAccountInformation()
    mainWindow.openMainWindow()

    sys.exit(app.exec())



from PyQt6.QtCore import QCoreApplication


def getLanguageNames() -> dict[str, str]:
    return {
        "en": QCoreApplication.translate("Language", "English"),
        "de": QCoreApplication.translate("Language", "German"),
        "nl": QCoreApplication.translate("Language", "Dutch"),
        "es": QCoreApplication.translate("Language", "Spanish"),
        "ru": QCoreApplication.translate("Language", "Russian"),
        "he": QCoreApplication.translate("Language", "Hebrew"),
    }



from typing import Optional, TYPE_CHECKING
import json
import os


if TYPE_CHECKING:
    from jdMinecraftLauncher.Environment import Environment


class MicrosoftSecrets:
    _instance: Optional["MicrosoftSecrets"] = None

    def __init__(self, env: "Environment") -> None:
        # In my opinion, it is not possible to hide the credentials from a person who really want it
        # This little "encryption" ist just to hide it from Bots

        self.client_id = ""
        self.secret = ""
        self.redirect_url = ""

        if os.path.isfile(os.path.join(env.dataDir, "secrets.json")):
            with open(os.path.join(env.dataDir, "secrets.json"), "r", encoding="utf-8") as f:
                self._json_data = json.load(f)
        elif os.path.isfile(os.path.join(env.currentDir, "secrets.json")):
            with open(os.path.join(env.currentDir, "secrets.json"), "r", encoding="utf-8") as f:
                self._json_data = json.load(f)
        else:
            pass

        self._decrypt("clientID", "client_id")
        self._decrypt("secret", "secret")
        self._decrypt("redirectURL", "redirect_url")

    def _decrypt(self, json_key: str, obj_key: str) -> None:
        if self._json_data[json_key] is None:
            setattr(self, obj_key, None)
            return
        if not self._json_data["encrypted"]:
            setattr(self, obj_key, self._json_data[json_key])
            return
        text = self._json_data[json_key][::-1]
        result = ""
        for c in text:
            result += chr(ord(c) - 5)
        setattr(self, obj_key, result)

    @classmethod
    def setup(cls, env: "Environment") -> None:
        cls._instance = cls(env)

    @classmethod
    def get_secrets(cls) -> "MicrosoftSecrets":
        if cls._instance is None:
            cls._instance = cls.__new__(cls)
        return cls._instance



from typing import cast, Optional, TYPE_CHECKING
from jdMinecraftLauncher.Profile import Profile
from PyQt6.QtCore import QCoreApplication
import traceback
import json
import sys
import os


if TYPE_CHECKING:
    from jdMinecraftLauncher.Environment import Environment


class ProfileCollection:
    def __init__(self, env: "Environment") -> None:
        self.profileList: list[Profile] = []
        self._loadError: str | None = None
        self.selectedProfile = ""
        self._env = env

    def loadProfiles(self) -> None:
        profileJSON = os.path.join(self._env.dataDir, "profiles.json")

        if not os.path.isfile(profileJSON):
            defaultProfile = Profile(QCoreApplication.translate("ProfileCollection", "Default"), self._env)
            self.selectedProfile = defaultProfile.id
            self.profileList.append(defaultProfile)
            return

        try:
            with open(profileJSON, "r", encoding="utf-8") as f:
                data = json.load(f)

            profileVersion = data.get("version", 0)

            self.selectedProfile = data.get("selectedProfile", "")

            for i in data["profileList"]:
                self.profileList.append(Profile.load(self._env, i, profileVersion))

            if profileVersion != 1:
                self.save()
        except Exception:
            self._loadError = traceback.format_exc()
            print(self._loadError, file=sys.stderr)

            if len(self.profileList) == 0:
                defaultProfile = Profile("Default", self._env)
                self.selectedProfile = defaultProfile.id
                self.profileList.append(defaultProfile)
                return

    def save(self) -> None:
        if self._env.args.dont_save_data:
            return

        data = {"selectedProfile": self.selectedProfile, "profileList": []}
        for i in self.profileList:
            cast(list[dict], data["profileList"]).append(i.toDict())
        data["version"] = 1  # type: ignore

        with open(os.path.join(self._env.dataDir, "profiles.json"), "w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=4)

    def getProfileByName(self, name: str) -> Optional[Profile]:
        for i in self.profileList:
            if i.name == name:
                return i
        return None

    def getProfileByID(self, profile_id: str) -> Optional[Profile]:
        for i in self.profileList:
            if i.id == profile_id:
                return i
        return None

    def getSelectedProfile(self) -> Profile:
        profile = self.getProfileByID(self.selectedProfile)
        if profile is not None:
            return profile
        else:
            return self.profileList[0]

    def getLoadError(self) -> Optional[str]:
        return self._loadError

    def removeProfileById(self, profileId: str) -> None:
        for pos, profile in enumerate(self.profileList):
            if profile.id == profileId:
                del self.profileList[pos]
                break

        if self.selectedProfile == profileId:
            self.selectedProfile = self.profileList[0].id



from ...ui_compiled.AboutTab import Ui_AboutTab
from PyQt6.QtWidgets import QWidget
from ...Functions import openFile
from typing import TYPE_CHECKING
import minecraft_launcher_lib
import webbrowser


if TYPE_CHECKING:
    from ...Environment import Environment


class AboutTab(QWidget, Ui_AboutTab):
    def __init__(self, env: "Environment") -> None:
        super().__init__()

        self.setupUi(self)

        self.versionLabel.setText(self.versionLabel.text().replace("{{version}}", env.launcherVersion))
        self.libLabel.setText(self.libLabel.text().replace("{{version}}", minecraft_launcher_lib.utils.get_library_version()))

        self.viewSourceButton.clicked.connect(lambda: webbrowser.open("https://codeberg.org/JakobDev/jdMinecraftLauncher"))
        self.openMinecraftDirectoryButton.clicked.connect(lambda: openFile(env.minecraftDir))
        self.openDataDirectoryButton.clicked.connect(lambda: openFile(env.dataDir))
